(****************************************************************************)
(*                                                                          *)
(* Ce fichier contient les fonctions utiles au calcule des BoundingBox      *)
(*                                                                          *)
(* Un BoundingBox se prsente sous la forme d'un 6-uplet de flottants.      *)
(* Les 3 premires coordonnes sont celles du coin infrieur du BoundingBox *)
(* dans le repre (X, Y, Z).                                                *)
(* Les 3 dernires sont celles du coin suprieur.                           *)
(*                                                                          *)
(****************************************************************************)



open Types
open List
open Triplet
open Parametres



(********************************************************)
(* Calcule la somme des lments flottants d'un tableau *)
(********************************************************)
(* Type : float array -> float                          *)
(********************************************************)
let sumTab = Array.fold_left (fun v s -> v +. s) 0.0



(****************************************************)
(* Calcule le minimum ou le maximum de 4 valeurs    *)
(****************************************************)
(* Type : float -> float -> float -> float -> float *)
(****************************************************)
let min_of_4 x1 x2 x3 x4 = min (min x1 x2) (min x3 x4)
let max_of_4 x1 x2 x3 x4 = max (max x1 x2) (max x3 x4)



(***********************************************************)
(* Calcule le BoundingBox de 2 objets                      *)
(***********************************************************)
(* Type : float * float * float * float * float * float -> *)
(*        float * float * float * float * float * float -> *)
(*        float * float * float * float * float * float    *)
(***********************************************************)
let bounding_fusion (minX1, minY1, minZ1, maxX1, maxY1, maxZ1) (minX2, minY2, minZ2, maxX2, maxY2, maxZ2) = (min minX1 minX2,
													     min minY1 minY2,
													     min minZ1 minZ2,
													     max maxX1 maxX2,
													     max maxY1 maxY2,
													     max maxZ1 maxZ2)


(***********************************************************************)
(* Met  jour la valeur de l ou t.(i) par le calcul de max(l, max-min) *)
(* Utilises pour le calcul des BB des Grids                           *)
(***********************************************************************)
(* Type : float ref -> float -> float -> unit                          *)
(***********************************************************************)
let max_diff_update_ref l ma mi = l := (max !l (ma -. mi))
(***********************************************************************)
(* Type : float array -> int -> float -> float -> unit                 *)
(***********************************************************************)
let max_diff_update_tab t i ma mi = t.(i) <- (max t.(i) (ma -. mi))



(*************************************************************************)
(* Permute les coordonnes du boundingBox pour passer du repre          *)
(* (X, Y, Z) au repre (axis1, axis2, ...)                               *)
(*************************************************************************)
(* Type : float * float * float * float * float * float -> axe -> axe -> *)
(*        float * float * float * float * float * float                  *)
(*************************************************************************)
let desordre (x1, y1, z1, x2, y2, z2) axis1 axis2 = match (axis1, axis2)
						    with
						      | (AxeX, AxeY) -> (x1, y1, z1, x2, y2, z2)
						      | (AxeX, AxeZ) -> (x1, z1, y1, x2, z2, y2)
						      | (AxeY, AxeX) -> (y1, x1, z1, y2, x2, z2)
						      | (AxeY, AxeZ) -> (y1, z1, x1, y2, z2, x2)
						      | (AxeZ, AxeX) -> (z1, x1, y1, z2, x2, y2)
						      | (AxeZ, AxeY) -> (z1, y1, x1, z2, y2, x2)
						      | _            -> raise (Erreur "BoundingBox : repre invalide")



(*************************************************************************)
(* Permute les coordonnes du boundingBox pour passer du repre          *)
(* (axis1, axis2, ...) au repre (X, Y, Z)                               *)
(*************************************************************************)
(* Type : float * float * float * float * float * float -> axe -> axe -> *)
(*        float * float * float * float * float * float                  *)
(*************************************************************************)
let reordre (x1, y1, z1, x2, y2, z2) axis1 axis2 = match (axis1, axis2)
						    with
						      | (AxeX, AxeY) -> (x1, y1, z1, x2, y2, z2)
						      | (AxeX, AxeZ) -> (x1, z1, y1, x2, z2, y2)
						      | (AxeY, AxeX) -> (y1, x1, z1, y2, x2, z2)
						      | (AxeY, AxeZ) -> (z1, x1, y1, z2, x2, y2)
						      | (AxeZ, AxeX) -> (y1, z1, x1, y2, z2, x2)
						      | (AxeZ, AxeY) -> (z1, y1, x1, z2, y2, x2)
						      | _            -> raise (Erreur "BoundingBox : repre invalide")



(***********************************************************)
(* Dtermine l'axe axis2 tel que (axis, axis2) soit direct *)
(***********************************************************)
(* Type : axe -> axe                                       *)
(***********************************************************)
let axe_suiv axis = match axis
		    with
		      | AxeX    -> AxeY
		      | AxeY    -> AxeZ
		      | AxeZ    -> AxeX
		      | AxeNone -> AxeNone



(**************************************************************)
(* Permute les coordonnes pour passer du repre              *)
(* (axis1, axis2, ...) au repre (X, Y, Z)                    *)
(* Ralise en quelque sorte l'inverse de la fonction desordre *)
(**************************************************************)
(* Type : float * float * float -> axe -> axe ->              *)
(*        float * float *float                                *)
(**************************************************************)
let permut (x, y, z) axis1 axis2 = match (axis1, axis2)
				   with
				     | (AxeX, AxeY) -> (x, y, z)
				     | (AxeX, AxeZ) -> (x, z, y)
				     | (AxeY, AxeX) -> (y, x, z)
				     | (AxeY, AxeZ) -> (z, x, y)
				     | (AxeZ, AxeX) -> (y, z, x)
				     | (AxeZ, AxeY) -> (z, y, x)
				     | (_, _)       -> raise (Erreur "BoundingBox : repre invalide")



(************************************************************)
(* Calcule la 1re partie de la translation sur un axe pour *)
(* passer d'un lment  un autre autre dans un Grid.       *)
(* align est l'alignement de l'objet sur l'axe, min et max  *)
(* les coordonnes sur cet axe du boundingbox de l'objet    *)
(************************************************************)
(* Type : align -> float -> float -> float                  *)
(************************************************************)
let sbt algn min max = (match algn
                        with
			  | Left   -> (max -. min) /. 2.0
			  | Middle -> 0.0
			  | Right  -> (min -. max) /. 2.0
		       ) -. ((min +. max) /. 2.0)



(************************************************************)
(* Calcule la 2nde partie de la translation sur un axe pour *)
(* passer d'un lment  un autre autre dans un Grid.       *)
(* align est l'alignement de l'objet sur l'axe, nouv et anc *)
(* les tailles sur cet axe du boundingbox de l'objet et de  *)
(* son prdcesseur.                                        *)
(************************************************************)
(* Type : align -> float -> float -> float                  *)
(************************************************************)
let sbt2 algn anc nouv = match algn
			 with
			   | Left   -> (anc -. nouv) /. 2.0
			   | Middle -> 0.0
			   | Right  -> (nouv -. anc) /. 2.0



(*************************************************************)
(* Calcule le boundingBox d'un Grid dans la base (X, Y, Z)   *)
(*  partir du centre du Grid center et des dimensions dim1, *)
(* dim2 et dim3 dans la base (axis1, axis2, ...)             *)
(*************************************************************)
(* Type : point -> float -> float -> float -> axe -> axe ->  *)
(*        float * float * float * float * float * float      *)
(*************************************************************)
let boundingBox_of center dim1 dim2 dim3 axis1 axis2 = let dims = permut (dim1 /. 2.0, dim2 /. 2.0, dim3 /. 2.0) axis1 axis2
						       in 
						       let (minX, minY, minZ) = diff_triplet center#getAll dims
						       and (maxX, maxY, maxZ) =  add_triplet center#getAll dims
						       in
							 (minX, minY, minZ, maxX, maxY, maxZ)



(***********************************************************)
(* Intervertit les valeurs min et max si min > max         *)
(***********************************************************)
(* Type : float * float * float * float * float * float -> *)
(*        float * float * float * float * float * float    *)
(***********************************************************)
let bonOrdre (minX, minY, minZ, maxX, maxY, maxZ) = (min minX maxX, min minY maxY, min minZ maxZ, max minX maxX, max minY maxY, max minZ maxZ)



(***********************************************************)
(* Dtermine le centre d'un boundingBox.                   *)
(***********************************************************)
(* Type : float * float * float * float * float * float -> *)
(*        float * float * float                            *)
(***********************************************************)
let middle (minX, minY, minZ, maxX, maxY, maxZ) = ((maxX +. minX) /. 2.0,
						   (maxY +. minY) /. 2.0,
						   (maxZ +. minZ) /. 2.0)


(*******************************************************************)
(* Fonction rcurrente pour calculer les BoundingBox de listes     *)
(* d'objets lst et d'un objet ayant le boundingbox bounding        *)
(*******************************************************************)
(* Type : list -> float * float * float * float * float * float -> *)
(*        float * float * float * float * float * float            *)
(*******************************************************************)
let rec _boundingList lst bounding = match lst
				      with
					| []   -> bounding
					| t::q -> let bounding2 = t#getBoundingBox ()
						  in
						    _boundingList q (bounding_fusion bounding bounding2)



(*******************************************************************)
(* Calcule le BoundingBox d'une liste d'objets lst                 *)
(*******************************************************************)
(* Type : list -> float * float * float * float * float * float -> *)
(*        float * float * float * float * float * float            *)
(*******************************************************************)
and  boundingList lst = match lst
			with
			  | []     -> (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
			  | hd::tl -> _boundingList tl (hd#getBoundingBox ())



(********************************************************)
(* Calcule le BoundingBox d'une scne scn               *)
(********************************************************)
(* Type : scene ->                                      *)
(*        float * float * float * float * float * float *)
(********************************************************)
let boundingScene scn = match scn#getContenu
			with
			  | []     -> (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
			  | hd::tl -> _boundingList tl (hd#getBoundingBox ())
