open Types
open Parametres
open Graph
open Boundingbox
open Primitive
open Polyline
open Polygon
open Triplet
open Box
open Font

class simplex (col:couleur) (wire:draw) (verticesVal : vertex list ref) (labelVal : string option) =
  object(self)
    inherit edge col wire Graph.dummyVertex Graph.dummyVertex as super
      
    (** Members *)

    (** List of primitives representing the connections*)
    val mutable primList = []

    (** Distance between a vertice and the primitive representing a connection *)
    val simplexDist = 0.3;

    (** A label that is displayed along with the simplex *)
    val mutable label = None

    val mutable center = Parametres.getOri ()

    (** Methods *)
    method setVertices newVertices = vertices <- newVertices;
                                     self#createPolys (self#deflate (Graph.toPointList vertices))

    (** Creates the list of primitives representing the connection*)
    method createPolys points= match points with
                                 | []                  -> ()
				 | p1::[]              -> ()
				 | p1::p2::[]          -> primList <- ((new polyline col wire (ref (p1::p2::[]))):>primitive) :: []
				 (*| p1::p2::p3::[]      -> primList <- ((new polygone col wire (ref (p1::p2::p3::[])) (ref []) false):>primitive) :: []
				 | p1::p2::p3::p4::[]  -> primList <- ((new polygone col wire (ref (p1::p2::p3::[])) (ref []) false):>primitive) ::
			                                              ((new polygone col wire (ref (p1::p2::p4::[])) (ref []) false):>primitive) ::
				                                      ((new polygone col wire (ref (p1::p3::p4::[])) (ref []) false):>primitive) ::
				       ((new polygone col wire (ref (p2::p3::p4::[])) (ref []) false):>primitive)::[]*)
				 | _                   -> (* Box represents center of simplex *)
                                                          let centerBox = new vertex center col (Some((new box col center wire None (new point 0.1 0.1 0.1)):>primitive)) "" None
							  (* Line from center to each vertex of the simplex *)
							  and lineList = List.map (fun p -> (new polyline col wire (ref (p::center::[])):>primitive)) points
							  in
							    primList <- (centerBox:> primitive)::lineList

    method drawSelf () = List.iter (fun prim -> prim#drawSelf ()) primList;
                         if (label <> None && (List.length vertices)> 1 && (not self#getDrawBlend)) then
			   begin
			     GlMat.translate3 center#getAll;
			     (* Draw label *)
			     (unSome label)#drawSelf ();
			     (* Reset position *)
                             GlMat.translate3 (Triplet.oppose_triplet center#getAll)
			   end


    method getBoundingBox () = Boundingbox.boundingList primList

    method setColor col = super#setColor col;
                          if (label <> None) then
			    (unSome label)#setColor col;
                          List.iter (fun prim -> prim#setColor col) primList

    (** Returns if the number of vertices does not exceed the maximum number of vertices *)
    method isLegal = true(*(List.length vertices)<5*)

    (** Recomputes the positions of the points such that the primitive representing the simplex will not reach till the points *)
    method deflate points = center <- Triplet.getCenter points;
                            if (List.length points>1) then
	                      begin
				List.map (fun p -> let np = ref ( p#getAll)
			                           and dist = Triplet.dist center p
				                   in
						     np := Triplet.diff_triplet !np center#getAll;
				                     np := Triplet.scaleTriplet !np (dist -. simplexDist);
						     np := Triplet.add_triplet !np center#getAll;
						     (triplet2point !np)
					 )
				         points
			      end
                            else
	                      points


    (** Update positions of vertices after placing *)
    method update = self#createPolys (self#deflate (Graph.toPointList vertices));
                    

    (** Writes the data of the edge to a channel *)
    method writeToChannel channel = output_string channel ("Simplex { ");
                                    if (label <> None) then
				      output_string channel ("Label \"" ^ (unSome label)#getText ^ "\" "); 
                                    output_string channel "VertexList [";
				    output_string channel (List.hd vertices)#getName;
				    List.iter (fun vert -> output_string channel ("," ^ vert#getName)) (List.tl vertices);
				    output_string channel "] ";
				    output_string channel ("Color <" ^ (string_of_float color#getRouge) ^ "," ^ (string_of_float color#getVert) ^ "," ^ (string_of_float color#getBleu) ^ "> }")
    
    initializer
      vertices <- !verticesVal;
      (* Create the primitives representing the simplex *)
      self#createPolys (self#deflate (Graph.toPointList vertices));
      (* Create label *)
      if (labelVal <> None) then
	let p = Parametres.getOri ()
	in
	  label <- Some (new label self#getColor self#getWire p ("  " ^ (unSome labelVal)))
  end

class complexGraph (col:couleur) (wire:draw) (vertices: vertex list ref) (simplexVal:simplex list ref)=
  object(self)
    inherit graph col wire vertices (ref []) false as super

    (** Members *)
    val mutable simplexs = !simplexVal;

    (** Methods *)

    (** Setters and Getters *)
    
    (** Return simplexs to overwrite method of graph to enable placing *)
    method getEdges = ref (List.map (fun conn -> (conn :> Graph.edge)) simplexs)

    method getSimplexs = ref simplexs
    method setSimplexs newSimplexs = simplexs <- !newSimplexs

    (** Normal method to return simplexs *)

    method drawSelf () = (* First draw connections, afterwards the vertices *)
                         List.iter (fun con -> con#draw ()) simplexs;
                         (* Second the vertices *)
                         List.iter (fun vert -> vert#draw ()) !vertices;

    (** Appends the data of the complex to an imoview file*)
    method append channel name = (*write intro*)
                                 output_string channel ("Scene " ^ name ^ " {\n");
                                 output_string channel "ComplexGraph {\n";
				 output_string channel "Color <0.0,1.0,0.0>\n";
			         output_string channel "VertexList[\n";
				 (List.hd !vertices)#writeToChannel channel;
				 List.iter (fun vert -> output_string channel ",\n";
			                                vert#writeToChannel channel;
					   )
			                   (List.tl !vertices);
				 output_string channel "]\n";
				 output_string channel "SimplexList[\n";
				 (List.hd simplexs)#writeToChannel channel;
				 List.iter (fun con -> output_string channel ",\n";
				                       con#writeToChannel channel;
					   )
				           (List.tl simplexs);
				 output_string channel ("]\n }\n}\nReplace{ Show " ^ name ^ "}\n")

    (** Writes the complex to a file *)
    method writeToFile fname = let out = open_out_gen [Open_creat;Open_trunc;Open_wronly] 0 fname 
                               in
				 self#append out "complex";
                                 close_out out

    initializer
      (* Set graph color to simplexes and vertices if not defined *)
      List.iter (fun vert  -> if (vert#getColor#isDefault) then vert#setColor col ) !vertices;
      List.iter (fun conn  -> if (conn#getColor#isDefault) then conn#setColor col ) simplexs;

      (* Set draw mode for vertices and edges if graph is blended *)
      (match wire with
	| Blend alpha -> List.iter (fun conn -> conn#setWire (Blend alpha)) simplexs;
	                 List.iter (fun vert  -> vert#setWire (Blend alpha)) !vertices
	| _           -> ()
      );

      (* Check if every connection has at most 4 edges *)
      (* List.iter (fun con -> if (not con#isLegal) then raise (Erreur("Illegal number of vertices in connection!")) )  simplexs *)

  end
