open Types
open Parametres
open Viewport
open Group
open Boundingbox
open Primitive

(** Defines a primitive that contains primitives itself. 
  Any class that inherits from this one must first call the function draw of the parent class
  to draw the content
@param col           Color of the primitive
@param position      Position of the primitive, not yet supported, use translation instead
@param wireVal       Defines how to draw the object
@param containerVal  Objects contained within the primitive

 *)
class virtual content (col:couleur) (position:point) (wireVal:draw) (containerVal: container option)=
  object (self)
    inherit primitive col wireVal as super
    
    (** contained primitives *)
    val mutable container = containerVal

    (** position of the primitive *)
    val mutable pos = position
   
    (** part of the primitive to be drawn*)
    val mutable sector = All

    (** scale within the primitive*)
    val mutable scale = 1.0
 
    (*-------*)
    (*Methods*)
    (*-------*)

    method setContainer newContainer = container <- Some(newContainer)
    method getContainer = unSome(container)
    
    method setPos newPos = pos <- newPos
    method getPos = pos

    method setContentMode newSector newScale newWire = sector <- newSector; scale <-newScale; wire <- newWire

    method setSector newSector = sector <-newSector
    method getSector = sector

    method setWire newWire = wire <- newWire
    method getWire = wire

    (** Draw primitive according draw-mode *)
    method drawWireSpecific wire = (* Set draw color *)
                                   self#setGLColor;
                                   match (wire) with
                                     | Wire when view#getPolygonMode = `fill      -> GlDraw.polygon_mode ~face:`both `line;
				                                                     GlDraw.cull_face `back;
										     Gl.disable `cull_face;
										     self#drawSpecific sector false;
										     GlDraw.cull_face `back;
										     Gl.enable `cull_face;
										     GlDraw.polygon_mode ~face:`both view#getPolygonMode;
				     | Mixed (down, up)                           -> (*Draw lower part specific*)
				                                                     self#setSector Inf;
				                                                     self#drawWireSpecific down;
					                                             self#setSector Sup;
			                                                             self#drawWireSpecific up

				     | Open                                       -> self#drawSpecific Bottom true

				     | Wire
				     | Close                                      -> self#drawSpecific sector (sector <> All || drawBlend )
				     | Blend alpha                                -> self#drawSpecific sector (sector <> All || drawBlend || alpha <> 1.0)
				                                                (*if drawBlend
				                                                then
				                                                  begin
										    let oldAlpha = self#getColor#getAlpha
										    in
										      self#getColor#setAlpha (oldAlpha*.alpha);
										      self#setGLColor;
										      self#drawSpecific sector true;
										      (* Reset color *)
										      self#getColor#setAlpha oldAlpha;
										  end
				                                                else
				                                                  begin
										    let mat = GlMat.get_matrix `modelview_matrix
										    and newWire = match sector
										      with
											| Bottom -> (* Impossible :@ *) Shadow
											| Inf    -> Mixed (Blend alpha,      Shadow)
											| Sup    -> Mixed (     Shadow, Blend alpha)
											| All    -> Blend alpha
										    in
										      self#setDrawBlend true;
										      view#addBlendObject (self:>primitive) (Raw.get_float (GlMat.to_raw mat) ~pos:11) mat
										  end*)
											   
				     | Shadow                                     -> ()

    (** Draw method, should not be overwritten by child-classes, specific draw-methods for those are called "drawSpecific".
        draws first the contained primitives and afterwards the primitive itself *)
    method draw () = (*Draw content of container, if available*)
                     if (container <> None) then
                       self#getContainer#draw ();

                     (* Call draw method of super class *)
                     super#draw ()

    (** This method is not used here -> call drawWireSpecific with corresponding drawMode *)
    method drawSelf () = self#drawWireSpecific self#getWire

    (** Draw primitive: cont_mode gives the sector to draw and the bool if the interior of the object should be drawn *)
    method virtual drawSpecific: cont_mode -> bool -> unit

    (** Returns bounding box of the container *)
    method getContainerBounds = match container with
                                  | None           -> (0.0, 0.0, 0.0, 0.0, 0.0, 0.0), ref nan, new point nan nan nan
				  | Some(cont)     -> let bb = boundingList cont#getContenue
				                      in
				                        cont#getPos#setAll (Triplet.oppose_triplet (middle bb));
				                        bb, ref cont#getScale, cont#getPos

  end
