open Viewport
open Types
open Triplet
open File
open Boundingbox

(***********************************************************)
(* Calcule le centre d'un BoundingBox                      *)
(***********************************************************)
(* Type : float * float * float * float * float * float -> *)
(*        float * float * float                            *)
(***********************************************************)
let milieu_bounding (minX, minY, minZ, maxX, maxY, maxZ) =
  let x = 0.5 *. (maxX +. minX)
  and y = 0.5 *. (maxY +. minY)
  and z = 0.5 *. (maxZ +. minZ)
  in
    (x, y, z)

(***************************************************************)
(* Calcule le rayon de la sphre circonscrite d'un BoundingBox *)
(***************************************************************)
(* Type : float * float * float * float * float * float ->     *)
(*        float                                                *)
(***************************************************************)
let rayon_bounding (minX, minY, minZ, maxX, maxY, maxZ) =
  let x = 0.5 *. (maxX -. minX)
  and y = 0.5 *. (maxY -. minY)
  and z = 0.5 *. (maxZ -. minZ)
  in
    sqrt (x*.x +. y*.y +. z*.z)


(*************************************************************)
(* Central class that manages all Non-GUI-aspects of imoview *)
(* should be kept as singleton                               *)
(*************************************************************)
class _core = 
  object (self)

    (*---------*)
    (* Members *)
    (*---------*)
    
    (* Viewport to display scenes *)
    val view = ref Viewport.view;

    (* List of all scenes with their names -> except anonymous scenes *)
    val mutable sceneList = (ref(Hashtbl.create 101):(string,scene) Hashtbl.t ref)

    (* Last added scene for the display of anonymous scenes*)
    val mutable lastScene = ref dummyscene

    (* List of named objects *)
    val mutable objectList = (ref(Hashtbl.create 101):(string,drawable) Hashtbl.t ref)

    (*List of commands to do*)
    val mutable command_todo = (new file:(commande file))

    (*List of commands that have been performed*)
    val mutable command_done = (new file:(commande file))

    (*State of imoview*)
    val mutable state = ref Stop

    (* State after execution of command *)
    val mutable nextState = ref (None:(anim * sens) option)

    (*Direction of execution of commands*)
    val mutable direction = ref Avant

    (* Intervale de temps entre 2 comandes *)
    val mutable cmdTimeout = ref 1

    (* Timer de l'animation *)
    val mutable timer = ref (None:GMain.Timeout.id option)
    
    (*---------*)
    (* Methods *)
    (*---------*)

    method setState newState = state <- ref newState
    method getState = !state
    
    method setNextState newState = nextState <- ref newState
    method getNextState = !nextState

    method setDir newDir = direction <- ref newDir
    method getDir = !direction
   
    method setTimer newTimer = timer <- ref newTimer
    method getTimer = !timer

    method getCommandToDo = command_todo
    method getCommandDone = command_done

    method setCommandTimeout newTimeout = cmdTimeout <- ref newTimeout
    method getCommandTimeout = !cmdTimeout

    (**************************)
    (* SCENE AND OBJECT LISTS *)
    (**************************)

    (************************************************)
    (* Ajoute la scne scn  la liste des scnes et *)
    (* ajoute la  commande qui affiche la scne.    *)
    (* Si la file d'attente des commandes est vide  *)
    (* et qu'aucune pause n'tait active, on lance  *)
    (* le gestionnaire de commandes.                *)
    (************************************************)
    (* Type : commande -> unit                      *)
    (************************************************)
    method prepareFirstScene = let bb = boundingScene !lastScene
                               in
                                 let (x, y, z) = milieu_bounding bb
				 in
				   !view#getMilieuDessin#setAll (x, y, z);
				   !view#setRayon (max 20.0 (4.0 *. (rayon_bounding bb) +. (sqrt (x*.x +. y*.y +. z*.z))));
				   !view#setZoom  (!view#getRayon);
				   flush stdout
        
    (* Adds a scene with a name to the list of scenes *)
    method addScene scn = (*Check if name already exists*)
                          try 
			    let tmp = Hashtbl.find !sceneList scn#getName in ()
                          with 
			    | Not_found  -> if (not scn#getAnonym) then
				              Hashtbl.add !sceneList scn#getName scn;
					    (*Prepare boundingbox if first scene*)
					    if (lastScene= ref dummyscene) then 
					      begin
						lastScene <- ref scn;
						self#prepareFirstScene
					      end
					    else
					      lastScene <- ref scn

			    | _          -> raise (Erreur ("la scene " ^ (scn#getName) ^ " est dja dfinie"))
				
    (* Adds a list of scenes to the list of scenes *)
    method addSceneList scnList = List.iter (self#addScene) scnList

    (* Removes, given its name, a scene from the list of scenes*)
    method removeScene name = try Hashtbl.remove !sceneList name 
                              with
				| Not_found -> raise( Erreur("la scene " ^ name ^ " n'est pas dfinie"))

    (*Removes a list of scenes, given their names*)
    method removeSceneList nameList = List.iter (self#removeScene) nameList

    (*Returns the scene belonging to a name*)
    method getScene name = try Hashtbl.find !sceneList name
                           with
				| Not_found -> raise( Erreur("la scene " ^ name ^ " n'est pas dfinie"))

    (* returns the last added scene *)
    method getLastScene = !lastScene

    method clearSceneList = Hashtbl.clear !sceneList;
                            lastScene <- ref dummyscene
    

    (* Adds an object with a name to the list of objects *)
    method addObject obj = (*Check if name already exists*)
                           try 
			     let temp = Hashtbl.find !objectList obj#getName in ()
                           with 
			     | Not_found  -> Hashtbl.add !objectList obj#getName obj
			     | _          -> raise( Erreur("l'objet " ^ obj#getName ^ " est dja dfini"))
				
    (* Adds a list of scenes to the list of scenes *)
    method addObjectList objList = List.iter (self#addObject) objList

    (* Removes, given its name, a scene from the list of scenes*)
    method removeObject name = try Hashtbl.remove !objectList name 
                               with
				 | Not_found -> raise (Erreur("l'objet " ^ name ^ " n'est pas dfini"))

    (*Removes a list of scenes, given their names*)
    method removeObjectList nameList = List.iter (self#removeObject) nameList

    (*Returns the object belonging to a name*)
    method getObject name = try Hashtbl.find !objectList name
                            with
			      | Not_found -> raise( Erreur("l'objet " ^ name ^ " n'est pas dfini"))

    method clearObjectList = Hashtbl.clear !objectList

    (* Deletes all scenes and prepares for the loading of a new imo-file*)
    method prepareNewLoad = (* Clear lists *)
                            self#clearSceneList;
                            self#clearObjectList;
			    !view#clearScenes;
                            
			    (* Clear commands *)
			    command_todo#clear ();
                            command_done#clear ();
                            
			    (* Set original point of view *)
			    !view#getVectX #setAll (1., 0.              , 0.              );
			    !view#getVectY #setAll (0., 0.86602540378445,-0.5             );
			    !view#getVectZ #setAll (0., 0.5             , 0.86602540378445);
			    !view#getCentre#setAll (0., 0.              , 0.              )
    
    (*********************************)
    (* MANIPULATION OF POINT OF VIEW *)
    (*********************************)

    (*****************************************************)
    (* Fonction  timer  qui  ralise  le  mouvement  mvt *)
    (* jusqu' son terme.                                *)
    (* oldTime   est  l'heure    laquelle   le  dernier *)
    (* raffraichissement du mouvement a t effectu.    *)
    (* next  est  l'argument    passer    la  fonction *)
    (* sceneTimer une fois le mouvement termin.         *)
    (*****************************************************)
    (* Type : mouvement -> float -> anim -> unit -> unit *)
    (*****************************************************)
    method move_func mvt oldTime next () = (* On supprime le timout pour viter les saturations *)
					GMain.Timeout.remove (unSome self#getTimer);
					self#setTimer None;

					let amplitude vitesse duree limite = min_triplet (mult_triplet (vitesse)#getAll duree) (limite)#getAll
					in
					  (* Si le mouvement n'est pas fini, on continue *)
					  if self#getState <> Mvt ||
					     (((mvt#transDelta)#getAll = (0.0, 0.0, 0.0)) &&
					      ((mvt#  rotDelta)#getAll = (0.0, 0.0, 0.0)) &&
					      ((mvt# spinDelta)#getAll = (0.0, 0.0, 0.0)))
					  then
					    begin
					      self#setState next;
					      if (self#checkNewState ()) then ();
					    end
					  else
					    begin
					      let actuTime = Unix.gettimeofday ()
					      in
					      (* Calcul de la dure depuis le dernier raffraichissement pour *)
					      (* connatre l'amplitude du morceau de mouvement  raliser.   *)
					      let deltaT = actuTime -. oldTime
					      in
					      (* Calcule des amplitudes de mouvement  raliser pour chaque type de mouvement *)
					      let trans = amplitude mvt#transSpeed deltaT mvt#transDelta
					      and rot   = amplitude mvt#  rotSpeed deltaT mvt#  rotDelta
					      and spin  = amplitude mvt# spinSpeed deltaT mvt# spinDelta
					      in
						(* Mise  jour des amplitudes restant  raliser *)
						(mvt#transDelta)#setAll (diff_triplet (mvt#transDelta)#getAll trans);
						(mvt#  rotDelta)#setAll (diff_triplet (mvt#  rotDelta)#getAll rot  );
						(mvt# spinDelta)#setAll (diff_triplet (mvt# spinDelta)#getAll spin );
						(* Ralisation des mouvements *)
						if self#getDir = Avant
						then
						  begin
						    !view#translation trans;
						    !view#rotation    rot;
						    !view#spin        spin (0.0, 0.0, 0.0)
						  end
						else
						  begin
						    !view#translation (mult_triplet trans (-1.0));
						    !view#rotation    (mult_triplet rot   (-1.0));
						    !view#spin        (mult_triplet spin  (-1.0)) (0.0, 0.0, 0.0)
						  end;
						if next <> Rapide
						then
						  !view#draw ();

						self#setTimer ( Some (GMain.Timeout.add ~ms:1
									                ~callback:(self#move_func mvt actuTime next)
								   ) );
					    end;
					  false

    (*********************)
    (* COMMAND EXECUTION *)
    (*********************)

    (*Adds a command to the list of commands -> till now no special treatment*)
    method addCommand cmd = command_todo#pushEnd cmd
    
    (* Determines if a command contains only temporary scenes, such that it can be deleted after execution *)
    method containsTempScene cmd = match cmd with 
                                     | Show(scnList)       -> (((List.length scnList)=1) && ((List.hd scnList)#getAnonym)) 
				     | Replace (hide,show) -> ((hide=None) && ((List.length show)=1) && ((List.hd show)#getAnonym) )
				     | _                   -> false
    

    (*Stops playing and removes all scenes from the display*)
    method doStop () = while not (command_done#isEmpty)
                       do
	                 command_todo#pushBegin (command_done#popBegin ())
                       done;
                       !view#clearScenes


    method animToString = !Parametres.ecrireStatut (match self#getState, self#getDir
						    with
						      | Lct   , Avant
						      | Step  , Avant   -> "Lecture"
						      | Lct   , Arriere
						      | Step  , Arriere -> "Lecture arrire"
						      | Rapide, Avant   -> "Avance rapide"
						      | Rapide, Arriere -> "Retour rapide"
						      | Stop  , _       -> "Stop"
						      | Tempo , _       -> "Pause"
						      | Fin   , _       -> "Fin"
						      | _               -> "Erreur"
						   )

    (*Execute a command depending on the state*)
    method exec_command cmd = let masque dir lst = List.iter (match dir
							      with
								| Avant   -> (fun scn -> !view#removeScene scn)
								| Arriere -> (fun scn -> !view#addScene scn)
							     )
                                                             lst
                              and sensOppose dir = match dir
				                   with
						     | Avant   -> Arriere
						     | Arriere -> Avant
			      and adaptReplace cmd dir = match dir with
				                           | Avant   -> let cmdDone = command_done#popBegin () in
							                command_done#pushBegin cmd;
						           | Arriere -> let cmdDone = command_todo#popEnd () in
							                command_todo#pushEnd cmd;
                              in
                                match cmd
				with
				  | Pause                     -> self#setState Tempo

				  | Show scnLst               -> masque (sensOppose self#getDir) scnLst; 

				  | Hide scnLst               -> masque             self#getDir  scnLst;

				  | Replace (hide, show)      -> (* Determine list of scenes that will be hidden to enable backward playback *)
                                                                 if (hide = None && (not (self#containsTempScene cmd))) then
								   begin
				     				     adaptReplace (Replace (Some(!view#getNamedScenes), show)) self#getDir
								   end;
				                                 if (hide <> None) then
								   masque             self#getDir  (unSome hide)
								 else
								   !view#clearScenes;
				                                 masque (sensOppose self#getDir) show				                                 
								 
				  | Command str               -> if self#getDir = Avant
				                                 then
					                           begin
								     print_endline str;
								     flush stdout;
								     ignore (Sys.command str)
								   end;

				  | View (viewr, centr, vert) -> (match self#getDir
								  with
								    | Avant   -> (* On fixe le centre de la vision *)
									         !view#getCentre#setAll (match centr
													 with
													   | AutoCenter -> !view#getMilieuDessin#getAll
													   | Center pt  -> pt#getAll
													);

									         (* On fixe le vecteur Z du repre de la camera *)
									         (match viewr
										   with
										     | Viewer pt  -> let vecteur = diff_triplet pt#getAll !view#getCentre#getAll
										                     in
										                       let norme = normale_triplet vecteur
												       in
													 !view#setZoom norme;
													 !view#getVectZ#setAll (div_triplet vecteur norme)
										     | AutoViewer -> !view#setZoom !view#getRayon;
											             !view#getVectZ#setAll (0.0, 0.5, 0.86602540378445)
										 );

									         (* On fixe la verticale du repre de la camera *)
										 let z = !view#getVectZ#getAll
										 in
										   let vect = (match vert
											       with
												 | AutoVertical -> if (colineaire (0.0, 1.0, 0.0) z)
												         	   then (0.0, 0.0, -1.0)
												                   else (0.0, 1.0,  0.0)
											 	 | Vertical pt  -> (if (colineaire pt#getAll z)
												                   then if (pt#getAll = (0.0, 1.0, 0.0))
														     then (0.0, 0.0, -1.0)
														     else (0.0, 1.0,  0.0)
												                   else pt#getAll
														  )
											      )
										   in
										     let sbt = normalise_triplet (diff_triplet vect (mult_triplet z (scalaire vect z)))
										     in
										       !view#getVectY#setAll sbt;

										       (* On fixe le verteur X du repre de la camera *)
										       !view#getVectX#setAll (prod_triplet sbt z)

								    | Arriere -> ()
								 );

				  | Movement mvt              -> (* On ajoute un timout ralisant le mouvement *)
				                                 let copie = new Types.mouvement
							         and nextState = self#getState
								 in
								   copie#setrotSpeed   (Oo.copy mvt#rotSpeed);
								   copie#settransSpeed (Oo.copy mvt#transSpeed);
								   copie#setspinSpeed  (Oo.copy mvt#spinSpeed);
								   copie#setrotDelta   (Oo.copy mvt#rotDelta);
								   copie#settransDelta (Oo.copy mvt#transDelta);
								   copie#setspinDelta  (Oo.copy mvt#spinDelta);
								   self#setState Mvt;
								   self#setTimer (Some (GMain.Timeout.add ~ms:1
											                 ~callback:(self#move_func copie (Unix.gettimeofday ()) nextState)
										      ) );



    method exec = let (cmdTodo, cmdDone) = (match self#getDir
					    with
					      | Avant   -> self#getCommandToDo, self#getCommandDone
					      | Arriere -> self#getCommandDone, self#getCommandToDo
					   )
			 in
			   match self#getState
			   with
			     | Stop   -> self#doStop ();
				         !view#draw ();

			     | Rapide -> while not (cmdTodo#isEmpty)
			                 do
				           let cmd = cmdTodo#popBegin ()
					   in
					     ignore (self#exec_command cmd);
					     cmdDone#pushBegin cmd;
			                 done;
				         !view#draw ();
					 self#setState Tempo

			     | Step   -> if not (cmdTodo#isEmpty)
			                 then
				           begin
					     let cmd = cmdTodo#popBegin ()
					     in 
					       let tempCommand = (self#containsTempScene cmd)
					       in
						 (* Remember command only if it contains at least one named scene *)			       
						 if (not tempCommand) then
						   cmdDone#pushBegin cmd;
						 self#exec_command cmd;
						 self#setState Tempo;
						 !view#draw ();
					   end
                 	                 else
				           begin  
                  			     if self#getDir = Avant
					     then
					       self#setState Fin
					     else
					       self#setState Tempo
					   end
					     
			     | Lct    -> if not (cmdTodo#isEmpty)
			                 then
				           begin
					     let cmd = cmdTodo#popBegin ()
					     in 
					       let tempCommand = (self#containsTempScene cmd)
					       in
						 (* Remember command only if it contains at least one named scene *)
						 if (not tempCommand) then
						   cmdDone#pushBegin cmd;
						 self#exec_command cmd;
						 !view#draw ();
					   end
 			                 else
				           begin 
					     if self#getDir = Avant
					     then
					       self#setState Fin
					     else
					       self#setState Tempo
					   end
					     
			     | Tempo  -> ()

			     | _      -> (* Ne doit pas arriver :$ *)
							()

    method checkNewState () = (* On supprime le timout pour viter les saturations *)
                              (*print_string "Next thing";
                              flush stdout;*)
                              if (self#getTimer <> None) then
				begin
				  GMain.Timeout.remove (unSome self#getTimer);
				  self#setTimer None;
				end;
			      (match self#getNextState
                              with
				| None            -> self#exec; 
				                     (match self#getState
				                      with
							| Lct   -> self#setTimer( Some (GMain.Timeout.add ~ms: self#getCommandTimeout
												          ~callback:(self#checkNewState )
									 		    ));
							           false

							| Rapide  
							| Step   
							| Fin
							| Stop
							| Tempo  -> self#animToString; false

							| Mvt    -> false
						     )

				| Some (nxt, dir) -> let oldState = self#getState and oldDir = self#getDir in
				                       self#setNextState None;
				                       self#setDir dir;
						       self#setState nxt;
						       self#animToString;
						       self#exec;

						       (*Restore old state*)
						       self#setState oldState;
						       self#setDir oldDir;

						       self#setTimer (Some (GMain.Timeout.add ~ms:10
									                      ~callback:(self#checkNewState )
									   ));
						       false )


  end

let core = new _core
