open Types
open Parametres
open Polyline
open Sphere
open Usefull
open Triplet
open Primitive
open Font

(** Default radius of a vertex if not given*)
let defaultVertexRad = 0.1

(** Implement vertex as a somehow reduces version of a sphere, stack and slices are set automatically to 6
    @param posVal     Position of the node
    @param objVal     Primitive that represents this vertex
    @param name       Name of the vertex
    @param labelVal   Label to display with this vertex, has the same color like the vertex
 *)
class vertex (posVal:point) (color:couleur) (objVal:primitive option) (nameVal:string) (labelVal:string option)=
  object (self)

    inherit primitive color Close as super

    (** Members *)
    val mutable pos=posVal
    
    (** For simple graph placer *)
    val mutable force = Parametres.getOri ()
    val mutable obj = objVal;

    (** index of a vertex *)
    val mutable index = 0;

    (** Label to display with this vertex *)
    val mutable label = None
 
    (** Methods *)
   
    method setForce newForce = force <- newForce; 
    method getForce = force
    method resetForce = force <- Parametres.getOri ()

    method setObject newObject = obj <- Some(newObject)
    method getObject = unSome(obj);

    method setPos newPos = pos <- newPos
    method getPos = pos

    method setIndex newIndex = index <- newIndex
    method getIndex = index

    method setColor newColor = super#setColor newColor;
                               (unSome obj)#setColor newColor

    (** Adds the force to the vertex and resets it *)
    method update = pos <- (Triplet.sumVec self#getPos force);
                    self#resetForce;
      
    method drawSelf () = (* Need to set correct point to draw sphere *)
                         GlMat.translate3 pos#getAll;
                         (unSome obj)#drawSelf ();
			 (* Draw label *)
			 if (label <> None && (not self#getDrawBlend)) then
			   (unSome label)#drawSelf ();
			 (* Reset position *)
                         GlMat.translate3 (Triplet.oppose_triplet pos#getAll);
			   

    method getBoundingBox = (unSome obj)#getBoundingBox

    (** Writes the data of the vertex to a channel *)
    method writeToChannel channel = output_string channel ("Vertex " ^ name ^ "{ ");
                                    if (label <> None) then
                                      output_string channel ("Label \"" ^ (unSome label)#getText ^ "\" ");
				    output_string channel ("Position <" ^ (string_of_float pos#getX) ^ "," ^ (string_of_float pos#getY) ^ "," ^ (string_of_float pos#getZ) ^ "> ");
				    output_string channel ("Color <" ^ (string_of_float color#getRouge) ^ "," ^ (string_of_float color#getVert) ^ "," ^ (string_of_float color#getBleu) ^ "> }")

    initializer
      self#setName nameVal;

      (* Set default object if none is defined *)
      if (obj = None) then
	obj <- Some ((new sphere color pos Close None defaultVertexRad 6 6):>Primitive.primitive);

      (* Set draw mode *)
      self#setWire (unSome obj)#getWire;
      
      (* Create label *)
      if (labelVal <> None) then
	let p = Parametres.getOri ()
	in
	  label <- Some (new label self#getColor self#getWire p (" " ^ (unSome labelVal)))
  end

(** Extracts the points from a list of vertices *)
let toPointList vertices = List.map (fun vert -> vert#getPos) vertices

(** List of all currently defined vertices, necessary when constructing the graph *)
let vertexList = (ref(Hashtbl.create 101):(string,vertex) Hashtbl.t ref)

let dummyVertex = new vertex Parametres.ori (new couleur nan nan nan nan) None "dummy" None

(** Implementation of edge as a reduces polyline with only 2 points 
    @param col   Color of the edge
    @param wire  DrawMode of the edge, only Blend will take effect
    @param v1    First vertex of the edge
    @param v2    Second vertex of the edge
*)
class edge (col:couleur) (wire:draw) (v1 : vertex) (v2:vertex) =
  object (self)
    inherit polyline col wire (ref ((v1#getPos)::(v2#getPos)::[])) as super

    (** Members *)
    (** List of vertices *)
    val mutable vertices = v1::v2::[]

    (** Weight of the vertex *)
    val mutable weight = 1.0

    (** Setters and getters *)
    method getV1 = List.nth vertices 0
    method getV2 = List.nth vertices 1

    method getVertices = ref vertices

    method getWeight = weight
    method setWeight newWeight = weight <- newWeight

    method update = (*print_string "Upped"; flush stdout;*)
                    self#setPoints ((self#getV1#getPos)::(self#getV2#getPos)::[])

    (** Writes the data of the edge to a channel *)
    method writeToChannel channel = output_string channel ("Edge { ");
                                    output_string channel ("V1 " ^ v1#getName ^ " V2 " ^ v2#getName ^ " ");
				    output_string channel ("Color <" ^ (string_of_float color#getRouge) ^ "," ^ (string_of_float color#getVert) ^ "," ^ (string_of_float color#getBleu) ^ "> }")

  end



class graph (col:couleur) (wire:draw) (verticesVal: vertex list ref) (edgesVal:edge list ref) (directedVal:bool)=
  object(self)
    inherit polyline col wire (ref []) as super

    (** Members *)
    
    (** Directed graph*)
    val mutable directed=directedVal

    (** Edge list *)
    val mutable edges = edgesVal

    (** Vertex list*)
    val mutable vertices = verticesVal

    (** Hashmap with name of vertices to find them faster *)
    val mutable vertexMap = vertexList

    (** Methods *)

    (* Setters and Getters *)
    method getVertices = vertices
    method setVertices newVertices = vertices <- newVertices

    method getDirected = directed
    
    (** Returns the vertex with the given name *)
    method getVertex name = Hashtbl.find !vertexMap name
	
    method getEdges = edges

    method drawSelf () = (* First draw edges, afterwards the vertices *)
                         List.iter (fun edg -> edg#draw ()) !edges;
                         (* Second the vertice *)
                         List.iter (fun vert -> vert#draw ()) !vertices;

    (** Updates the graph after some changes, e.g. when a placer was applied.
	
	Current Updates:
	----------------

	Updates the position of the vertices and sets the list of points to enable the computation of the boundingbox
        by the inherited function of polyline
    *)
    method update () = self#setPoints ( Usefull.fold (fun vert      -> vert#getPos)
					             (fun head tail -> head::tail )
						     []
						     !vertices
				      );

    (** Writes the whole graph to a file*)
    method writeToFile fname = let channel = open_out_gen [Open_creat;Open_trunc;Open_wronly] 0 fname 
                               in
				 (*write intro*)
				 output_string channel "Scene graph {\n";
                                 output_string channel "Graph {\n";
				 output_string channel "VertexList[\n";
				 (List.hd !vertices)#writeToChannel channel;
				 List.iter (fun vert -> output_string channel ",\n";
				                        vert#writeToChannel channel;
					   )
				           (List.tl !vertices);
				 output_string channel "]\n";
				 output_string channel "EdgeList[\n";
				 (List.hd !edges)#writeToChannel channel;
				 List.iter (fun edg -> output_string channel ",\n";
				                       edg#writeToChannel channel;
					   )
				           (List.tl !edges);
                                 output_string channel "]\n}\n}\nShow graph";
				 flush channel;
				 close_out channel


    initializer 
      (* Set color to default graph color if not yet defined in vertices or edges *)
      List.iter (fun edg  -> if (edg#getColor#isDefault) then edg#setColor col ) !edges;
      List.iter (fun vert -> if (vert#getColor#isDefault) then vert#setColor col ) !vertices;

      (* Set draw mode for vertices and edges if graph is blended *)
      (match wire with
	| Blend alpha -> List.iter (fun vert -> vert#setWire (Blend alpha)) !vertices;
	                 List.iter (fun edg  -> edg#setWire (Blend alpha)) !vertices
	| _           -> ()
      );
	
      self#update ();

      (* Set the indices of the vertices *)
      let i = ref 0
      in
	List.iter (fun vert -> vert#setIndex !i; 
	                      i:= !i + 1
		 )
	         !vertices;
	
     (* reset vertex list to enable loading of next graph *)
     (*vertexList := ((Hashtbl.create 101):(string,vertex) Hashtbl.t)*)
  end

(** Graph Utils *)

(** Prints an array to the screen *)
let print_it arr = for i = 0 to ((Array.length arr) - 1 ) do
                     for j = 0 to ((Array.length arr.(i)) - 1 ) do
		       print_float arr.(i).(j); print_string " ";
		     done;
		     print_string "\n";
		   done;
		   flush stdout

(** Computes the shortest path between each pair of vertices of a graph *)
let shortest_path gr = (* Create the adjacency matrix *)
                       let n = List.length !(gr#getVertices)
		       and changed = ref true 
		       and max = (float_of_int max_int)
		       in
		         let adj = Array.make_matrix n n max
			 and dist = Array.make_matrix n n max
			 in
			   List.iter (fun edg -> if ((List.length !(edg#getVertices) )> 1) then
			                           let v1 = edg#getV1#getIndex and v2 = edg#getV2#getIndex
			                           in
						     adj.(v1).(v2) <- edg#getWeight;
						     adj.(v2).(v1) <- edg#getWeight;
						     dist.(v1).(v2) <- edg#getWeight;
						     dist.(v2).(v1) <- edg#getWeight;
				     ) 
			             !(gr#getEdges);

			   (* Initialize diagonal with zero *)
			   for i = 0 to (n-1) do
			     adj.(i).(i) <- 0.0;
			     dist.(i).(i) <- 0.0;
			   done;
			   (* Iterate until steady state is reached *)
			   while (!changed) do
			     changed := false;
			     for i = 0 to (n-1) do
			       for j = i+1 to (n-1) do
				 (* Get minimum of row *)
				 let min = ref max
				 in
				   for k = 0 to (n-1) do
				     if ((dist.(i).(k) +. adj.(k).(j) < !min)) then
				       min := dist.(i).(k) +. adj.(k).(j);
				   done;
				   if (!min <> dist.(i).(j) ) then
				     begin
				       dist.(i).(j) <- !min;
				       dist.(j).(i) <- !min;
				       changed := true;
				     end;
			       done
			     done
			   done;
			   dist



			 
