open Types
open Parametres
open Graph
open Viewport

(** Show information while placing *)
let showDetails = true

class virtual graphPlacer (graphVal : graph) =
  object(self)
    
    (** Members *)
    (** The graph to place *)
    val mutable theGraph = graphVal

    (** Vertices for which the placing should be performed, by default all vertices *)
    val mutable vertices = graphVal#getVertices

    (** Edges between those vertices, by default all edges *)
    val mutable edges = graphVal#getEdges

    (** Bool indicating if placer is available (check on startup )*)
    val mutable canPlace = true;

    (** Just do determine running time *)
    val mutable startTime = 0.0;

    (** Methods *)
    
    (* Setters and Getters *)
    method setGraph newGraph = theGraph <- newGraph
    method getGraph = theGraph

    (** Graph placing *)

    (* Initialize internal data structures *)
    method virtual init: unit -> unit

    (** Update position of each node *)
    method virtual update: unit -> unit

    (** Termination condition, placement is being performed while termination condition is not yet fulfilled *)
    method virtual finished : unit -> bool

    (** Places all vertices that have no positions on random positions
	@param dim if set to 3 random positions also random z-position is set
    *)
    method placeRandom dim = (* Determine approximated area of random placing for vertices without position *)
                             let range = (sqrt (float_of_int (List.length !vertices))) *. 3.0
			     in 
			       let xRange = range and yRange = range and zRange = range 
			       (* Determine the center of all points *)
			       and ori = Triplet.sumList (List.map (fun vert-> vert#getPos) !vertices) 
			       in 
				 (* Determine base of placement*)
				 ori#updateX (-. xRange /. 2.0);
				 ori#updateY (-. yRange /. 2.0);
				 ori#updateZ (-. zRange /. 2.0);
				 
				 (* Determine random position for all vertices that are left*)
				 List.iter (fun vert -> if (vert#getPos#isDefault) then
		                                          begin
							    vert#getPos#setX (ori#getX +. Random.float xRange);
							    vert#getPos#setY (ori#getY +. Random.float yRange);
							    if (dim = 3) then
							      vert#getPos#setZ (ori#getZ +. Random.float zRange)
							    else 
							      vert#getPos#setZ 0.0
							  end
					   ) !vertices;
				 
				 (* Update positions of vertices *)
				 List.iter (fun edg -> edg#update) !edges

    (** Perform placing *)
    method place = if (canPlace) then
	             begin
		       (* To check running time *)
		       startTime <- Unix.gettimeofday ();
		       (* Init placer*)
		       self#init ();
                       if (showDetails) then 
			 begin
			   print_string "Starting placing \n"; flush stdout
			 end;
		       (* Do first placing, since OCAML does only know while loops *)
		       self#update ();
                       while (not (self#finished ()) ) do
			 self#update ()
                       done;
		       print_string "Finished placing\nThe process took ";
		       print_float ((Unix.gettimeofday ()) -. startTime );
		       print_string " s."; flush stdout		       
		     end

    (** Shows the steps of the placing algorithm *)
    method showPlacing = if (canPlace) then
	                   begin
			     (* Init placer*)
			     self#init ();
                             (* Create new scene with graph as content *)
                             if (showDetails) then 
                               begin
				 print_string "Starting placing \n"; flush stdout
			       end;
                             let scn = new scene 
                             in 
                               scn#setContenu ((theGraph :> Types.drawable)::[]);
			       (* Start placing in seperate thread *)
                               let t = (GMain.Timeout.add ~ms:100
                                	   ~callback:(fun () -> (* Update positions *)
					                        self#update ();
					                        (* Draw scene *)
					                        view#draw ();
								(* Check if termination condition is fulfilled *)
					                        if ( self#finished () ) then
								  begin
								    print_string "Finished placing\n"; flush stdout;
								    false
								  end
								else
								  true										
				                     )
				       ) in ()
			   end

  end

class simplePlacer (graphVal : graph) (placeAll:bool) (dimVal:int)=
  object(self)
    inherit graphPlacer graphVal as super

    (** Members *)
    (** Attractive force between connected vertices *)
    val attForce = 0.2

    (** Repulsive force between any two vertices. The equilibrium distance between two connected nodes can be computed as
    square root of attForce/repForce *)
    val repForce = 0.2

    (** Resolution of the force calculation, integration time steps *)
    val resolution = 0.1

    (** Sum of all forces, to determine when to terminate*)
    val forceSum = Parametres.getOri ()

    (** Treshhold of forces per vertice. If the sum of forces is below this value the placing is finished *)
    val treshold = 0.002

    (** Dimension in which the placer works, currently only working with 2 and 3*)
    val dim = dimVal

    (** Place vertices without positions randomly *)
    method init () = (* Determine vertices that need to be placed *)
                     if (not placeAll) then
		       vertices <- ref (List.filter (fun vert -> vert#getPos#isDefault ) !vertices);
                     super#placeRandom dim

    method update () = let repell vert = List.iter (fun v2 -> if (vert <> v2 ) then
                                                                let d = Triplet.dist vert#getPos v2#getPos and move = vert#getForce 
                                                                in
								  move#updateX ((vert#getPos#getX -. v2#getPos#getX) *. resolution *. repForce /. (d**2.0));
								  move#updateY ((vert#getPos#getY -. v2#getPos#getY) *. resolution *. repForce /. (d**2.0));
								  move#updateZ ((vert#getPos#getZ -. v2#getPos#getZ) *. resolution *. repForce /. (d**2.0));
						    ) !vertices
                       and attract edge = let center = Triplet.getCenter (Graph.toPointList !(edge#getVertices))
			                  in
					    List.iter (fun vert -> let d = Triplet.dist vert#getPos center and force = Parametres.getOri ()
					                           in
								     force#setX ((center#getX -. vert#getPos#getX) *. resolution *. attForce *. d);
					                             force#setY ((center#getY -. vert#getPos#getY) *. resolution *. attForce *. d);
								     force#setZ ((center#getZ -. vert#getPos#getZ) *. resolution *. attForce *. d);
								     vert#getForce#updateAll force#getAll;
						      ) 
					              !(edge#getVertices)
                       in
			 List.iter (repell) !vertices;
			 List.iter (attract) !edges;
			 (* Compute the sum of all forces to determine if termination condition is fulfilled *)
			 self#sumForces;
			 List.iter (fun vert -> vert#update) !vertices;
			 List.iter (fun edg -> edg#update) !edges

    (** Computes the sum of all forces *)
    method sumForces = List.iter (fun vert -> forceSum#updateAll (Triplet.absTriplet (vert#getForce#getAll))) !vertices;
                       if (showDetails) then 
                         begin
			   print_string "Force: "; print_float (Triplet.vecLength forceSum); print_string "\n"; flush stdout
			 end

    method finished () = let d =(Triplet.vecLength forceSum) 
                         in
                           (* Reset sum of forces *)
                           forceSum#updateAll (Triplet.oppose_triplet forceSum#getAll);
			   (* Check if sum of forces is below treshold*)
			   if (d < (treshold *. (float_of_int (List.length !edges)))) then 
			     begin
			       print_string "Finished"; flush stdout;
			     end;
			   (d < (treshold *. (float_of_int (List.length !edges))))

  end

(** Execute a command and returns a outchannel to write into its inchannel and an inchannel to read its output 
   @return (process id, inchannel, outchannel) *)
let getProcess command args = (* Create two pipes to read and write into the process channels *)
                              let (inbound_in,inbound_out) = Unix.pipe () and (outbound_in,outbound_out) = Unix.pipe ()
			      in
				(* Create channels to be returned as Pervasive.channel *)
			        let inchannel = Unix.in_channel_of_descr outbound_in 
				and outchannel = Unix.out_channel_of_descr inbound_out
				in 
				  (* Create process *)
				  let p = Unix.create_process command args inbound_in outbound_out (Unix.descr_of_out_channel stdout);
				  in
				    (p,inchannel,outchannel)

	     

(** Sign that neato has finished computation *)
let finishExp = Str.regexp "}"

(** Parent of all graphPlacers that use the dot-format with the respective methods to convert into 
    this format and parse the obtained results
    @param command Command to start the placer
*)
class virtual dotPlacer (graphVal:graph) (commandVal:string) =
  object(self)
    inherit graphPlacer graphVal as super

    (** Members *)
    (** Command to start the placer*)
    val command = commandVal

    (** Methods *)
      
    method init () = (* Check if the placer is available on the system *)
                     print_string "Checking placer version:\n"; flush stdout;
                     let exitSignal = Unix.system (command ^ " -V")
		     in
		       if (exitSignal = Unix.WEXITED(127)) then
			 begin
			   canPlace <- false;
			   print_string ("Could not find placer " ^ command ^ " in PATH."); flush stdout
			 end

    (** Creates a string representation of the contained graph in the dot-format 
	@params placeAll   if all nodes should be placed, even if a position is given
	@params stereo     writes a graph with z-coordinates
    *)
    method dottify          placeAll stereo   = let return = ref "" 
                                                in 
                                                  let writePos pos = return := !return ^ "pos = \"" ^ (string_of_float pos#getX) ^ "," ^ (string_of_float pos#getY);
								     if (stereo) then
								       return := !return ^ "," ^ (string_of_float pos#getZ);
						                     (* Check if node should be placed although a position is given *)
						                     if (not placeAll) then
								       return := !return ^ "!";
								     return := !return ^ "\""
						  in 
						    (* Writes a vertex to the output-string *)
						    let writeVertex vert = if (not vert#getPos#isDefault) then
				                                             begin
									       return := !return ^ vert#getName ^ " [ ";
									       writePos vert#getPos;
									       return := !return ^ " ]; \n";
									     end
						    (* Write an edge to the output-string*)
						    and writeEdge edg = return := !return ^ edg#getV1#getName;
				                                        if (theGraph#getDirected) then
									  return := !return ^ " -> "
									else
									  return := !return ^ " -- ";
									return := !return ^ edg#getV2#getName ^ "; \n"
						    in
						      (* Write header *)
						      if (theGraph#getDirected) then
							return := "digraph G {\n"
						      else 
							return := "graph G {\n";
						      (* Properties of the graph *)
						      if (stereo) then
							return := !return ^ "dim = 3;\n";
						      (* Vertices *)
						      List.iter (writeVertex) !(theGraph#getVertices);
						      (* Edges *)
						      List.iter (writeEdge) !(theGraph#getEdges);
						      (* Trailer *)
						      return := !return ^ "}\n";
						      !return

    (** Parses the obtained results from the placer back into the graph 
	@params result String-representation of the computed graph
	@params stereo z-coordinates are also given and should therefore be parsed 
    *)						       
    method parseResult result stereo = 	   let posExp  = Str.regexp ".*pos=\""
					   and vertExp = Str.regexp ".*-[->].*"
					   and stringArr = Str.split (Str.regexp "\n") result
					   and trim s = Str.replace_first (Str.regexp "^[ \t\n]+") "" s
					   in
                                             (* Returns the in the string saved position *)	
                                             let getPos2D posString = let redString = Str.replace_first posExp "" posString 
				                                      in 
					                                let kommaPos = (String.index redString ',')
									and quotePos = (String.index redString '"')
									in
									  let xPos = (float_of_string (String.sub redString 0 kommaPos))/.72.0
									  and yPos = (float_of_string (String.sub redString (kommaPos+1) (quotePos-kommaPos-1)))/.72.0
									  in
							                    let  p = new point xPos yPos 0.0
								            in 
									      p
					     and getPos3D posString = let redString = Str.replace_first posExp "" posString
					                              in
					                                let firstComma = String.index redString ','
									and quotePos = String.index redString '"'
									in
									  let secondComma = String.index_from redString (firstComma+1) ',' 
									  in
									    let xPos = (float_of_string (String.sub redString 0 firstComma))/.72.0
							                    and yPos = (float_of_string (String.sub redString (firstComma+1) (secondComma-firstComma-1)))/.72.0
							                    and zPos = (float_of_string (String.sub redString (secondComma+1) (quotePos-secondComma-1)))/.72.0
							                    in
							                      let  p = new point xPos yPos zPos
								              in 
										p

					     (* Returns the to the line corresponding vertex *)
					     and getVertex vertString = let inString = (trim vertString)
				                                        in
				                                          let vertName = String.sub inString 0 (String.index inString ' ')
									  in 
									    theGraph#getVertex vertName
					     (* Returns true if a line contains data about a vertex *)
					     and containsVertex testStr = ((Str.string_match posExp testStr 0) && (not (Str.string_match vertExp testStr 0)))
					     in
					       (* Analyse data *)
					       List.iter (fun line -> if (containsVertex line) then
					                                 begin
									   (* print_string line; flush stdout;*)
									   if (stereo) then 
									     (getVertex line)#setPos (getPos3D line)
									   else
									     (getVertex line)#setPos (getPos2D line);
									 end;
							 )
					                 stringArr;

					       (* Update the position of the vertices *)
					       List.iter (fun edg -> edg#update) !(theGraph#getEdges);

					       (* Update the graph itself *)
					       theGraph#update ();

  end


(** GraphPlacer that uses neato from the GraphViz-package as placer
    @param graphVal the graph on which the placing should be performed
    @param placeAll place all vertices even if positions are given (given positions will be used as initial positions)
*)
class neato (graphVal:graph) (placeAll:bool) =
  object(self)
    inherit dotPlacer (graphVal) "neato" as super
    
    (** Members *)    
    val mutable finish = false

    method update () = let graphString = self#dottify placeAll false
                       and (p,inchannel,outchannel) = getProcess "neato" [|""|]
                       and temp = ref ""
                       and resString = ref ""
                       in
			 (* Send graph to neato *)
			 output_string outchannel graphString;
                         flush outchannel;
                         (* Wait until computation is finished *)
                         while (not (Str.string_match finishExp !temp 0)) do
			   temp := input_line inchannel;
			   resString := !resString ^ !temp ^ "\n";
			 done;
			 (* Parse results *)
			 self#parseResult !resString false;
			 finish <- true;
		   

    method finished () = finish
  end


(** GraphPlacer that uses dot from the GraphViz-package as placer
    @param graphVal the graph on which the placing should be performed
    @param placeAll place all vertices even if positions are given (given positions will be used as initial positions)
*)
class dot (graphVal:graph) (placeAll:bool) =
  object(self)
    inherit dotPlacer (graphVal) "dot" as super
    
    (** Members *)    
    val mutable finish = false

    method update () = let graphString = self#dottify placeAll false
                       and (p,inchannel,outchannel) = getProcess "dot" [|""|]
                       and temp = ref ""
                       and resString = ref ""
                       in
			 (* Send graph to neato *)
			 output_string outchannel graphString;
                         flush outchannel;
                         (* Wait until computation is finished *)
                         while (not (Str.string_match finishExp !temp 0)) do
			   temp := input_line inchannel;
			   resString := !resString ^ !temp ^ "\n";
			 done;
			 (* Parse results *)
			 self#parseResult !resString false;
			 finish <- true;
		   

    method finished () = finish
  end


(** GraphPlacer that uses twopi from the GraphViz-package as placer
    @param graphVal the graph on which the placing should be performed
    @param placeAll place all vertices even if positions are given (given positions will be used as initial positions)
*)
class twopi (graphVal:graph) (placeAll:bool) =
  object(self)
    inherit dotPlacer (graphVal) "twopi" as super
    
    (** Members *)    
    val mutable finish = false

    method update () = let graphString = self#dottify placeAll false
                       and (p,inchannel,outchannel) = getProcess "twopi" [|""|]
                       and temp = ref ""
                       and resString = ref ""
                       in
			 (* Send graph to neato *)
			 output_string outchannel graphString;
                         flush outchannel;
                         (* Wait until computation is finished *)
                         while (not (Str.string_match finishExp !temp 0)) do
			   temp := input_line inchannel;
			   resString := !resString ^ !temp ^ "\n";
			 done;
			 (* Parse results *)
			 self#parseResult !resString false;
			 finish <- true;
		   

    method finished () = finish
  end


(** GraphPlacer that uses circo from the GraphViz-package as placer
    @param graphVal the graph on which the placing should be performed
    @param placeAll place all vertices even if positions are given (given positions will be used as initial positions)
*)
class circo (graphVal:graph) (placeAll:bool) =
  object(self)
    inherit dotPlacer (graphVal) "circo" as super
    
    (** Members *)    
    val mutable finish = false

    method update () = let graphString = self#dottify placeAll false
                       and (p,inchannel,outchannel) = getProcess "circo" [|""|]
                       and temp = ref ""
                       and resString = ref ""
                       in
			 (* Send graph to neato *)
			 output_string outchannel graphString;
                         flush outchannel;
                         (* Wait until computation is finished *)
                         while (not (Str.string_match finishExp !temp 0)) do
			   temp := input_line inchannel;
			   resString := !resString ^ !temp ^ "\n";
			 done;
			 (* Parse results *)
			 self#parseResult !resString false;
			 finish <- true;
		   

    method finished () = finish
  end


(** GraphPlacer that uses fdp from the GraphViz-package as placer
    @param graphVal the graph on which the placing should be performed
    @param placeAll place all vertices even if positions are given (given positions will be used as initial positions)
*)
class fdp (graphVal:graph) (placeAll:bool) =
  object(self)
    inherit dotPlacer (graphVal) "fdp" as super
    
    (** Members *)    
    val mutable finish = false

    method update () = let graphString = self#dottify placeAll false
                       and (p,inchannel,outchannel) = getProcess "fdp" [|""|]
                       and temp = ref ""
                       and resString = ref ""
                       in
			 (* Send graph to neato *)
			 output_string outchannel graphString;
                         flush outchannel;
                         (* Wait until computation is finished *)
                         while (not (Str.string_match finishExp !temp 0)) do
			   temp := input_line inchannel;
			   resString := !resString ^ !temp ^ "\n";
			 done;
			 (* Parse results *)
			 self#parseResult !resString false;
			 finish <- true;
		   

    method finished () = finish
  end
