open Types
open Parametres
open GraphPlacer
open Graph
open ComplexGraph

(** Graph drawing after Kamada & Kawai 1989 *)
class kkPlacer (graphVal:graph) (placeAll:bool) (dimVal:int) =
  object(self)
    inherit graphPlacer graphVal as super
    
    (** Array of vertices, for better performance *)
    val mutable tripletList = (Array.of_list !(graphVal#getVertices) )

    val mutable paths = (Graph.shortest_path graphVal)

    val mutable forceArr = Array.make (List.length !(graphVal#getVertices)) 0.0

    (** Approximated length of the cube containing the graph *)
    val l0 = 200

    (** Optimal length of an edge in the graph*)
    val mutable optLength = 5.0

    (** spring constant*)
    val k0 = 1.0

    (** Force at which a node is considered as being placed *)
    val mutable eps = 0.01

    (** Maximum of iterations to place a single node*)
    val maxIter = 50

    (** Maximum number of repeated placings for a node to avoid deadlock in local minima*)
    val maxIterGlobal = 20

    (** Node that was placed the last iteration*)
    val mutable lastPlaced = 0
    
    (** Number of times the last node was already placed *)
    val mutable repeatCount = 0;

    (** Dimension in which the placer works, currently only working with 2 and 3 *)
    val dim = dimVal

    (** Number of global iterations *)
    val mutable iterCount = 0

    (** Methods *)

    method init () = (* Determine vertices that need to be placed *)
                     if (not placeAll) then
		       vertices <- ref (List.filter (fun vert -> vert#getPos#isDefault ) !vertices);

                     (* Place vertices without position randomly*)
                     self#placeRandom dim;
		 
		     (* Calculate forces a first time *)
		     self#calcForces
    
    (* The following two functions might be replaced by some matrix, but it would be rather big ... *)

    (** ? *)
    method l i j = optLength *. paths.(i).(j)

    (** Force of a spring between vertex i and vertex j*)
    method k i j = k0 /. ( (paths.(i).(j)) ** 2.0 )

    (** Calculates the forces on a single vertex *)
    method calcForce index = let n = (Array.length tripletList) 
                             in
			       for j = 0 to (n-1) do
				 if (index<>j) then
				   begin
				     let dist = Triplet.dist tripletList.(index)#getPos tripletList.(j)#getPos
				     and p1 = tripletList.(index)#getPos and p2 = tripletList.(j)#getPos
				     in
				       tripletList.(index)#getForce#updateX ( (self#k index j) *. ((p1#getX -. p2#getX) -. ((self#l index j)*.(p1#getX -. p2#getX)) /. dist));
				       tripletList.(index)#getForce#updateY ( (self#k index j) *. ((p1#getY -. p2#getY) -. ((self#l index j)*.(p1#getY -. p2#getY)) /. dist));
				       tripletList.(index)#getForce#updateZ ( (self#k index j) *. ((p1#getZ -. p2#getZ) -. ((self#l index j)*.(p1#getZ -. p2#getZ)) /. dist));
				   end
			       done;
                               (*print_int index; print_string " ";
                               ipletList.(index)#getForce#print;*)
                               forceArr.(index) <- Triplet.vecLength tripletList.(index)#getForce

    (** Adds or removes the impact of a single vertex to all other vertices, more effecient than calculating these values every iteration again.
	@param dir If - 1.0 force of vertex index will be removed from force array, else it will be added
    *)
    method updateForce dir index = let n = (Array.length tripletList) 
                                   in
				     for j = 0 to (n-1) do
				       if (index<>j) then
					 begin
					   let dist = Triplet.dist tripletList.(index)#getPos tripletList.(j)#getPos
					   and p1 = tripletList.(index)#getPos and p2 = tripletList.(j)#getPos
					   and newForce = Parametres.getOri ()
					   in
					     newForce#updateX ( dir *. (self#k j index) *. ((p2#getX -. p1#getX) -. ((self#l j index)*.(p2#getX -. p1#getX)) /. dist));
					     newForce#updateY ( dir *. (self#k j index) *. ((p2#getY -. p1#getY) -. ((self#l j index)*.(p2#getY -. p1#getY)) /. dist));
					     newForce#updateZ ( dir *. (self#k j index) *. ((p2#getZ -. p1#getZ) -. ((self#l j index)*.(p2#getZ -. p1#getZ)) /. dist));
					     tripletList.(j)#getForce#updateAll newForce#getAll;
					     forceArr.(j) <- Triplet.vecLength tripletList.(j)#getForce;
					     (*int_int j; print_string " ";
					     tripletList.(j)#getForce#print;*)
					 end
				     done;


    (** Calculates the forces on each node in an initial step *)
    method calcForces = let n = (Array.length tripletList) 
                        in
			  for i = 0 to (n-1) do
			    self#calcForce i
			  done

    (** Perform an iteration step for a single node *)
    method update () = let n = (Array.length tripletList) 
                       and index = ref 0
                       and iter = ref 0
                       in
                         iterCount <- iterCount + 1;
                         (* Refreshs maxForceInd -> finds index with the maximum of force to be moved in the next step *)
                         let maxForce = for i = 0 to (n-1) do
			                  if (forceArr.(i)>forceArr.(!index)) then
					    index := i
			                  done
			 in
			   (* Check if node was already placed the last iteration *)
			   if (!index = lastPlaced) then
			     repeatCount <- repeatCount + 1
                           else
			     begin
			       repeatCount <- 0;
			       lastPlaced <- !index
			     end;

			   (* Remove the impact of the current vertex from the forces on the other vertices *)
			   self#updateForce (-. 1.0) !index;
			   while ((forceArr.(!index)>eps) && (!iter < maxIter)) do
			     (* Calculate coefficients (values of the Jacobi-Matrix) *)
			     let axx = ref 0.0 and axy = ref 0.0 and axz = ref 0.0
			     and ayx = ref 0.0 and ayy = ref 0.0 and ayz = ref 0.0
			     and azx = ref 0.0 and azy = ref 0.0 and azz = ref 0.0
			     and d1 = -. tripletList.(!index)#getForce#getX
			     and d2 = -. tripletList.(!index)#getForce#getY
			     and d3 = -. tripletList.(!index)#getForce#getZ
			     and p1 = tripletList.(!index)#getPos
			     in
			       (* Determine values of the Jacobi-Matrix*)
			       for j = 0 to (n-1) do
				 if (j <> !index) then
				   begin
				     (* Divisor of the coefficients *)
				     let div = (Triplet.dist tripletList.(!index)#getPos tripletList.(j)#getPos)**3.0
				     and p2 = tripletList.(j)#getPos
				     and lij = (self#l !index j)
				     and kij = (self#k !index j)
				     in
				       axx := !axx +. kij *. (1.0 -. lij *. ((p1#getY -. p2#getY) ** 2.0 +. (p1#getZ -. p2#getZ)**2.0 ) /. div );
				       axy := !axy +. kij *. (lij *. ((p1#getX -. p2#getX) *. (p1#getY -. p2#getY)) /. div );
				       axz := !axz +. kij *. (lij *. ((p1#getX -. p2#getX) *. (p1#getZ -. p2#getZ)) /. div );
				       ayx := !axy;
				       ayy := !ayy +. kij *. (1.0 -. lij *. ((p1#getX -. p2#getX) ** 2.0 +. (p1#getZ -. p2#getZ)**2.0 ) /. div );
				       ayz := !ayz +. kij *. (lij *. ((p1#getY -. p2#getY) *. (p1#getZ -. p2#getZ)) /. div );
				       azx := !axz;
				       azy := !ayz;
				       azz := !azz +. kij *. (1.0 -. lij *. ((p1#getX -. p2#getX) ** 2.0 +. (p1#getY -. p2#getY)**2.0 ) /. div );
				   end;
			       done;
			       (* Solving the implied linear system ...*)
			       let div = (!axz ** 2.0) *. !ayy -. 2.0 *. !axy *. !axz *. !ayz +. (!axy ** 2.0) *. !azz +. !axx *. ((!ayz ** 2.0) -. !ayy *. !azz )
			       in
			         let dx = ((!ayz ** 2.0) *. d1 -. !ayy *. !azz *. d1 +. !axy *. !azz *. d2 +. !axz *. !ayy *. d3 -. !ayz *. (!axz *. d2 +. !axy *. d3)) /. div
			         and dy = (!axy *. !azz *. d1 +. (!axz ** 2.0) *. d2  -. !axx *. !azz *. d2 +. !axx *. !ayz *. d3 -. !axz *. (!ayz *. d1 +. !axy *. d3)) /. div
				 and dz = (!axz *. !ayy *. d1 -. !axy *. !ayz *. d1 -. !axy *. !axz *. d2 +. !axx *. !ayz *. d2 +. (!axy ** 2.0) *. d3 -. !axx *. !ayy *. d3) /. div
				 in
				   (* Update positions*)
				   p1#updateAll (dx,dy,dz);
				   (* Recalculate forces *)
				   (*List.iter (fun vert -> vert#resetForce) !vertices;*)
				   tripletList.(!index)#resetForce;
				   self#calcForce !index;
				   iter := !iter + 1
			   done;
			   (* Update positions *)
			   List.iter (fun edg -> edg#update) !edges;
			   (* Calculate the impact of the new position of the moved vertex on the other vertices *)
			   (*List.iter (fun vert -> vert#resetForce) !vertices;*)
			   self#updateForce 1.0 !index;
			   (*self#calcForces*)
			   
    method finished () = let sum = ref 0.0 
                         and n = (List.length !vertices)
                         in
			   for i = 0 to (n - 1) do
			     sum := !sum +. forceArr.(i);
			   done;
                           if ((iterCount mod 100) = 1 ) then
			     begin
                               print_string "Force: "; print_float !sum; print_string "\n";flush stdout
			     end;
			   (((eps *. (float_of_int n)) > !sum) || (repeatCount>maxIterGlobal))

  end

class kkComplexPlacer (graphVal:complexGraph) (placeAll:bool) (dimVal:int)=
  object(self)
    inherit kkPlacer (graphVal :> graph) placeAll dimVal as super
      

    method init () = (* Determine vertices that need to be placed *)
                     if (not placeAll) then
		       vertices <- ref (List.filter (fun vert -> vert#getPos#isDefault ) !vertices);

                     (* Place nodes *)
                     self#placeRandom dim;

                     (* Change some parameters *)
                     optLength <- 5.0;
      
                     (* For any simplex, add a vertex in the center with a connection to every vertex of the simplex *)
                     (* Need only simplex with more than one vertice *)
		     let newVertices = graphVal#getVertices
		     and newSimplexs = graphVal#getSimplexs
		     in
		       List.iter (fun con -> let verts = con#getVertices
	                                     in
					       if (List.length !verts > 2) then
						 begin
						   (* Construct center to connect with other vertices *)
						   let center = new vertex (Triplet.getCenter (List.map (fun vert -> vert#getPos) !verts)) (con#getColor) (None) "temp" (Some "temp")
						   in
						     (* New vertex *)
						     (*newVertices := center :: !newVertices;*)
						     (* New edges *)
						     (*List.iter (fun vert -> let newCon = new simplex (con#getColor) (con#getWire) (ref (center::vert::[]))
					                                      in
						                                newSimplexs := newCon::!newSimplexs
						                 )
						                 !verts;*)
						     (* Simplex between every pair of vertices *)
						     for i=0 to ((List.length !verts) - 2) do
						       for j = i+1 to ((List.length !verts) - 1) do
							 let newCon = new simplex (con#getColor) (con#getWire) (ref ((List.nth !verts i)::(List.nth !verts j)::[])) None
							 in
							   newSimplexs := newCon::!newSimplexs;
							   (* Decrease weight to increase attraction*)
							   newCon#setWeight 1.0
						       done
						     done
						 end
				 )
			         !(graphVal#getSimplexs);
		       
		       (* Initialize internal data structures *)
		       tripletList <- (Array.of_list !(newVertices) );
		       
		       (* To compute paths matrix *)
		       let tempGraph = new graph (theGraph#getColor) (theGraph#getWire) (newVertices) (ref (List.map (fun con -> (con:> Graph.edge))  !newSimplexs)) false
		       in
			 paths <- (Graph.shortest_path tempGraph);

			 forceArr <- (Array.make (List.length !(newVertices)) 0.0);

			 (* Recalculate forces *)
			 List.iter (fun vert -> vert#resetForce) !vertices;
			 self#calcForces

  end
