(************************************************)
(*                                              *)
(* Ce fichier contient des fonction utiles pour *)
(* travailler sur des triplets de floattants.   *)
(*                                              *)
(************************************************)

open Usefull
open Types

(****************************************************)
(* Multiplie les coordonnes d'un triplet (x, y, z) *)
(* par un coefficiant c.                            *)
(****************************************************)
(* Type : float * float * float -> float ->         *)
(*        float * float * float                     *)
(****************************************************)
let mult_triplet (x, y, z) c = (c*.x, c*.y, c*.z)



(*************************************************)
(* Divise les coordonnes d'un triplet (x, y, z) *)
(* par un coefficiant c.                         *)
(*************************************************)
(* Type : float * float * float -> float ->      *)
(*        float * float * float                  *)
(*************************************************)
let div_triplet (x, y, z) c = (x/.c, y/.c, z/.c)



(************************************************************)
(* Additionne deux triplets (x1, y1, z1) et (x2, y2, z2).   *)
(************************************************************)
(* Type : float * float * float -> float * float * float -> *)
(*        float * float * float                             *)
(************************************************************)
let add_triplet (x1, y1, z1) (x2, y2, z2) = (x1+.x2, y1+.y2, z1+.z2)



(************************************************************)
(* Soustrait deux triplets (x1, y1, z1) et (x2, y2, z2).    *)
(************************************************************)
(* Type : float * float * float -> float * float * float -> *)
(*        float * float * float                             *)
(************************************************************)
let diff_triplet (x1, y1, z1) (x2, y2, z2) = (x1-.x2, y1-.y2, z1-.z2)

(*********************************************************)
(* Retourne l'oppos d'un triplet.                       *)
(*********************************************************)
(* Type : float * float * float -> float * float * float *)
(*********************************************************)
let oppose_triplet (x, y, z) = (-.x, -.y, -.z)


(***********************************)
(* Retourne l'oppos d'un point.   *)
(***********************************)
(* Type : point -> point           *)
(***********************************)
let oppose_point p = new point (-.p#getX) (-.p#getY) (-.p#getZ)


(*******************************************************)
(* Retourne la 1re coordonne d'un triplet (x, y, z). *)
(*******************************************************)
(* Type : float * float * float -> float               *)
(*******************************************************)
let abscisse (x, _, _) = x



(*******************************************************)
(* Retourne la 2me coordonne d'un triplet (x, y, z). *)
(*******************************************************)
(* Type : float * float * float -> float               *)
(*******************************************************)
let ordonnee (_, y, _) = y



(*******************************************************)
(* Retourne la 3me coordonne d'un triplet (x, y, z). *)
(*******************************************************)
(* Type : float * float * float -> float               *)
(*******************************************************)
let cote (_, _, z) = z



(********************************************************************)
(* Calcule le carr de la norme euclidienne d'un vecteur (x, y, z). *)
(********************************************************************)
(* Type : float * float * float -> float                            *)
(********************************************************************)
let normale_carre_triplet (x, y, z) = x*.x +. y*.y +. z*.z



(********************************************************)
(* Calcule la norme euclidienne d'un vecteur (x, y, z). *)
(********************************************************)
(* Type : float * float * float -> float                *)
(********************************************************)
let normale_triplet (x, y, z) = sqrt (x*.x +. y*.y +. z*.z)



(***********************************************************)
(* Calcule le vecteur unitaire de mme sens que (x, y, z). *)
(***********************************************************)
(* Type : float * float * float -> float * float * float   *)
(***********************************************************)
let normalise_triplet t = div_triplet t (normale_triplet t)



(************************************************************)
(* Calcule le produit scalaire de 2 vecteurs (x1, y1, z1)   *)
(* et (x2, y2, z2).                                         *)
(************************************************************)
(* Type : float * float * float -> float * float * float -> *)
(*        float                                             *)
(************************************************************)
let scalaire (x1, y1, z1) (x2, y2, z2) = x1*.x2 +. y1*.y2 +. z1*.z2



(************************************************************)
(* Calcule le produit vectoriel de 2 vecteurs (x1, y1, z1)  *)
(* et (x2, y2, z2).                                         *)
(************************************************************)
(* Type : float * float * float -> float * float * float -> *)
(*        float * float * float                             *)
(************************************************************)
let prod_triplet (x1, y1, z1) (x2, y2, z2) =
  (y1*.z2 -. z1*.y2,
   z1*.x2 -. x1*.z2,
   x1*.y2 -. y1*.x2)



(************************************************************)
(* Teste si 2 vecteures (x1, y1, z1) et (x2, y2, z2) sont   *)
(* colinaires.                                             *)
(************************************************************)
(* Type : float * float * float -> float * float * float -> *)
(*        bool                                              *)
(************************************************************)
let colineaire t1 t2 = ((prod_triplet t1 t2) == (0.0, 0.0, 0.0))



(****************************************************************)
(* Calcule le triplet compos pour chaque coordonne de la plus *)
(* petites valeurs de 2 vecteurs (x1, y1, z1) et (x2, y2, z2).  *)
(****************************************************************)
(* Type : float * float * float -> float * float * float ->     *)
(*        float * float * float                                 *)
(****************************************************************)
let min_triplet (x1, y1, z1) (x2, y2, z2) = (min x1 x2, min y1 y2, min z1 z2)



(*****************************************)
(* Convertit un triplet en point.        *)
(*****************************************)
(* Type : float * float * float -> point *)
(*****************************************)
let triplet2point (x, y, z) = new Types.point x y z

let absTriplet (x,y,z) = (abs_float x, abs_float y, abs_float z)

let lengthTriplet (x,y,z) = sqrt(x *.x +. y *. y +. z *. z)

let scaleTriplet (x,y,z) l = let scale = l/.(lengthTriplet (x,y,z)) in (scale *. x, scale *. y, scale *. z)

(*returns the length of a vector*)
let vecLength vec = vec#getX*.vec#getX+.vec#getY*.vec#getY+.vec#getZ*.vec#getZ;;

(*Maps a function to each coordinate of a point*)
let mapPoint f p = new point (f p#getX) (f p#getY) (f p#getZ);;

(*Maps a function to each coordinate of a point*)
let mapPoints f p1 p2 = new point (f p1#getX p2#getX) (f p1#getY p2#getY) (f p1#getZ p2#getZ);;

(*returns the sum of two points*)
let sumVec p1 p2= mapPoints (fun x y->x+.y) p1 p2;;

(*returns the difference between two points*)
let difVec p1 p2= mapPoints (fun x y->x-.y) p1 p2;;

(*Scales a vector to a given length*)
let scaleVec vec l = let scale = l/.sqrt(vecLength vec) in new point (scale *. vec#getX) (scale *. vec#getY) (scale *. vec#getZ);;

(*sums a list to a list of lists using sumLists*)
let rec addValToList list value = List.map (fun x-> sumVec x value) list;;

(*Scales a vector to length 1*)
let unifyVec vec = scaleVec vec 1.0;;

(*Multiplies a vec with a number*)
let multVec vec num = mapPoint (fun x -> x*.num) vec

(** Distance between two points *)
let dist p1 p2 = sqrt((p1#getX -. p2#getX) ** 2.0 +. (p1#getY -. p2#getY) ** 2.0 +. (p1#getZ -. p2#getZ) ** 2.0)

(** Returns a vector with each coordinate being the absolute value of this in the given vector*)
let abs_vec vec = mapPoint (abs_float) vec

(** Returns the sum of a list of vectors*)
let sumList l = let summe = Parametres.getOri ()
                in 
                  multVec (summe) (1.0 /. (float_of_int (List.length l))) 

(** Returns the center of a list of points *)
let getCenter points = let center = Usefull.fold (fun x -> x) 
                                                 (fun p1 p2 -> p2#updateAll p1#getAll;
						               p2
						 )
						 (Parametres.getOri ())
						 points
                       in
			 multVec center (1.0 /. (float_of_int (List.length points)))
			   

(***********************************)
(* Vector manipulation for display *)
(***********************************)



(*returns the normale of a polygone whose points lie in a single plane*)
let getPolygoneNormale pointList = (*Creates a vector that is orthogonal to a given polygone *)
				   let getNormale p1 p2 = unifyVec (new point (p1#getZ*.p2#getY -. p1#getY*.p2#getZ) (p1#getX *. p2#getZ -. p1#getZ*.p2#getX) 
                                                                   (p1#getY*.p2#getX -. p1#getX*.p2#getY)) in
				     match pointList with 
					|p1::p2::p3::tail -> getNormale (difVec p1 p2) (difVec p3 p2)
					|_ -> failwith "Not enough points to determine normale";;

(*rotates vec the given angle about axis, axis need to be a vector of length 1*)
let rotateVec vec axis angle =
  let sina = (sin angle) and
     cosa = (cos angle) in
	match (axis,vec) with
	  | (p,v) -> new point ((cosa +. p#getX**2.0*.(1.0-.cosa))*.v#getX  
				 +. (p#getX*.p#getY*.(1.0-.cosa)-.p#getZ*.sina)*.v#getY  
		    		 +. (p#getX*.p#getZ*.(1.0-.cosa)+.p#getY*.sina)*.v#getZ)
	                       ((p#getX*.p#getY*.(1.0-.cosa)+.p#getZ*.sina)*.v#getX 
		                 +. (cosa+.p#getY**2.0*.(1.0-.cosa))*.v#getY 
		                 +. (p#getY*.p#getZ*.(1.0-.cosa)-.p#getX*.sina)*.v#getZ)
		               ((p#getX*.p#getZ*.(1.0-.cosa)-.p#getY*.sina)*.v#getX 
		                 +. (p#getY*.p#getZ*.(1.0-.cosa)+.p#getX*.sina)*.v#getY 
			         +. (cosa+.p#getZ**2.0*.(1.0-.cosa))*.v#getZ);;

(*****************************************************)
(* Mise  jour de v1 et v2 aprs la rotation d'angle *)
(* theta autour de v1 vectoriel v2                   *)
(*****************************************************)
(* Type : float * float * float ->                   *)
(*        float * float * float -> unit              *)
(*****************************************************)
let rotate_vect v1 v2 theta =
  let coord1 = v1#getAll
  and coord2 = v2#getAll
  and cos_theta = cos theta
  and sin_theta = sin theta
  in
    v1#setAll (add_triplet (mult_triplet coord1 cos_theta) (mult_triplet coord2 sin_theta));
    v2#setAll (add_triplet (mult_triplet coord2 cos_theta) (mult_triplet coord1 (-.sin_theta)))



(****************************************************)
(* Multiplie 2 matrices 4*4 a et b.                 *)
(****************************************************)
(* Type : float array array -> float array array -> *)
(*        float array array                         *)
(****************************************************)
let mult_matrice a b = let res = [| Array.make 4 0.;
				    Array.make 4 0.;
				    Array.make 4 0.;
				    Array.make 4 0.
				 |]
		       in
			 for i = 0 to 3
			 do
			   for j = 0 to 3
			   do
			     res.(i).(j) <- (a.(i).(0) *. b.(0).(j)) +. (a.(i).(1) *. b.(1).(j)) +. (a.(i).(2) *. b.(2).(j))+. (a.(i).(3) *. b.(3).(j))
			   done
			 done;
			 res
