(*******************************************************************)
(*                                                                 *)
(* Ce fichier contient l'ensemble des types utiliss dans Imoview. *)
(*                                                                 *)
(*******************************************************************)




(******************)
(* Les exceptions *)
(******************)
exception ParseError of string
exception Erreur     of string
exception Help
exception APas

let unSome x = match x
	       with
		 | None   -> raise APas
		 | Some y -> y

(**************************************)
(* Types de fichiers images supports *)
(**************************************)
type format = 
   | Pdf
   | PsVect
   | EpsVect
   | Bmp
   | Gif
   | Jpeg
   | Ppm
   | Png
   | Ps



(***********************)
(* Sens de l'animation *)
(***********************)
type sens =
  | Avant
  | Arriere



(***********************)
(* Etat de l'animation *)
(***********************)
type anim =
  | Fin    (* Arrt car la pile des commandes est vide *)
  | Lct    (* Lecture -> excution d'un autre type de commande *)
  | Rapide (* Avance ou Retour rapide *)
  | Stop   (* Arrt de l'animation *)
  | Tempo  (* Pause -> attente d'un appui sur ESPACE *)
  | Step   (* N'effectue qu'une seule commande *)
  | Mvt    (* L'animation est en cour (le thread est dj lanc) *)



(*********************************)
(* Etat d'un bouton de la souris *)
(*********************************)
type mouseState =
  | DOWN
  | UP



(*****************************)
(* Type d'entre des donnes *)
(*****************************)
type input =
  | File
  | Stdin



(*********************************************************)
(* Type d'affichage des sphres, botes et ellipsodes : *)
(*  - All    -> entier                                   *)
(*  - Inf    -> moitie infrieure                       *)
(*  - Sup    -> moitie suprieure                       *)
(*  - Bottom -> Fond                                     *)
(*********************************************************)
type cont_mode =
  | All
  | Inf
  | Sup
  | Bottom



(**************************************)
(* Etat de chaque bouton de la souris *)
(**************************************)
class mouseButton =
object
  (* Position x en coordonne de fentre lors du clic *)
  val mutable x     = 0.0
  (* Position y en coordonne de fentre lors du clic *)
  val mutable y     = 0.0
  (* Etat du clic : DOWN ou UP *)
  val mutable state = UP
  (* Dtermine si la touche SHIFT tait enfonce lors du du clic *)
  val mutable shift = false
  (* Dtermine si la touche CTRL tait enfonce lors du du clic *)
  val mutable ctrl  = false

  (* Mthodes pour accder et modifier les donnes *)
  method getX              = x
  method getY              = y
  method setX newX         = x <- newX
  method setY newY         = y <- newY
  method getState          = state
  method setState newState = state <- newState
  method getShift          = shift
  method setShift newShift = shift <- newShift
  method getCtrl           = shift
  method setCtrl   newCtrl = ctrl  <- newCtrl
end



(*********)
(* Point *)
(*********)
class point (initX:float) (initY:float) (initZ:float) =
object
  (* Coordonne suivant X *)
  val mutable x = initX
  (* Coordonne suivant Y *)
  val mutable y = initY
  (* Coordonne suivant Z *)
  val mutable z = initZ

  (* Mthodes pour accder et modifier les coordonnes une par une *)
  method getX                   = x
  method getY                   = y
  method getZ                   = z
  method setX newX              = x <- newX
  method setY newY              = y <- newY
  method setZ newZ              = z <- newZ

  (* Mthodes pour accder et modifier toutes les coordonnes *)
  method getAll                 = (x, y, z)
  method setAll (xx, yy, zz)    = x <- xx;
				  y <- yy;
				  z <- zz

  (* Mthodes pour incrmenter les coordonnes *)
  method updateX n              = x <- x +. n
  method updateY n              = y <- y +. n
  method updateZ n              = z <- z +. n
  method updateAll (xx, yy, zz) = x <- x +. xx;
				  y <- y +. yy;
				  z <- z +. zz
  (* Mthodes opposes *)
  method oppose () =    x<-(-.x);
                        y<-(-.y);
                        z<-(-.z)
                          
  (* Dtermine  il s'agit d'une valeur par dfaut (non fixe) *)
  method isDefault = (x == nan || x == nan || x == nan)

  (* Print point to the screen*)
  method print = print_float x; print_string "/"; print_float y; print_string "/"; print_float z; print_string "\n"; flush stdout
end



(****************************************************************)
(* Couleur                                                      *)
(* Une couleur par dfaut comporte une composante nan.          *)
(* Une couleur valide a ses compossantes dans l'intervale [0;1] *)
(****************************************************************)
class couleur (initR:float) (initV:float) (initB:float) (initAlpha:float) =
object
  (* Composante rouge *)
  val mutable rouge = initR
  (* Composante verte *)
  val mutable vert  = initV
  (* Composante bleu *)
  val mutable bleu  = initB
  (* Transparence *)
  val mutable alpha = initAlpha

  (* Mthodes pour accder et modifier les composantes de la couleur *)
  method getRouge                = rouge
  method getVert                 = vert
  method getBleu                 = bleu
  method getAlpha                = alpha
  method getAll                  = (rouge, vert, bleu, alpha)
  method setRouge newRouge       = rouge <- newRouge
  method setVert newVert         = vert <- newVert
  method setBleu newBleu         = bleu <- newBleu
  method setAlpha newAlpha       = alpha <- newAlpha
  method setAll (rg, vt, bl, al) = rouge <- rg;
				   vert  <- vt;
				   bleu  <- bl;
				   alpha <- al

  (* Dtermine  il s'agit d'une couleur par dfaut (non fixe) *)
  method isDefault               = (rouge == nan || vert == nan || bleu == nan || alpha == nan)

  method print = print_string "<"; 
                 print_float rouge; print_string ",";
		 print_float vert; print_string ",";
		 print_float bleu; print_string ">";
		 flush stdout
end


(********)
(* Axes *)
(********)
type axe =
  | AxeX
  | AxeY
  | AxeZ
  | AxeNone (* AxeNone sert juste dans le parser de valeur par dfaut *)



(*******************)
(* Transformations *)
(*******************)
type transformation =
  | Rotation    of axe * float (* axe  angle *)
  | Scale       of point       (* taux *)
  | Translation of point       (* vecteur *)

(**************)
(* Alignement *)
(**************)
type align =
  | Left
  | Middle
  | Right

class virtual drawable =
  object(self)
    
    (*---------*)
    (* Members *)
    (*---------*)
    val mutable name=""
 
    (*---------*)
    (* Methods *)
    (*---------*)    
    method setName newName = name <- newName
    method getName = name

    (*Draw object*)
    method virtual draw : unit -> unit

    (****************)
    (* BOUNDING BOX *)
    (****************)
    method virtual getBoundingBox : unit -> (float*float*float*float*float*float)
    
  end


(************************)
(* Forme des Containers *)
(************************)
type typeContainer =
  | ContSphere of float ref
  | ContEllipsoide of point
  | ContBox of point



(***************************************)
(* Attribut d'affichage des Containers *)
(***************************************)
type draw =
  | Blend of float
  | Mixed of draw * draw (* Inf Sup *)
  | Shadow
  | Open
  | Wire
  | Close

(********************)
(* Vue de la camera *)
(********************)
type viewer =
    (* Vue automatique *)
  | AutoViewer
    (* Vue personnalise *)
  | Viewer of point



(********************)
(* Centre de la vue *)
(********************)
type center =
    (* Centre automatique *)
  | AutoCenter
    (* Centre personnalis *)
  | Center of point



(***********************)
(* Verticale de la vue *)
(***********************)
type vertical =
    (* Verticale automatique *)
  | AutoVertical
    (* Verticale personnalis *)
  | Vertical of point



(**************************)
(* Mouvement de la camera *)
(**************************)
class mouvement =
object
  (* Vitesse de translation *)
  val mutable _transSpeed = new point 0.0 0.0 0.0
  (* Amplitude de translation *)
  val mutable _transDelta = new point 0.0 0.0 0.0
  (* Vitesse de rotation par rapport au centre de la vue *)
  val mutable   _rotSpeed = new point 0.0 0.0 0.0
  (* Amplitude de rotation par rapport au centre de la vue *)
  val mutable   _rotDelta = new point 0.0 0.0 0.0
  (* Vitesse de rotation de la scne sur elle-mme *)
  val mutable  _spinSpeed = new point 0.0 0.0 0.0
  (* Amplitude de rotation de la scne sur elle-mme *)
  val mutable  _spinDelta = new point 0.0 0.0 0.0

  (* Mthodes pour accder aux attributs du mouvement *)
  method transSpeed = _transSpeed
  method   rotSpeed =   _rotSpeed
  method  spinSpeed =  _spinSpeed
  method transDelta = _transDelta
  method   rotDelta =   _rotDelta
  method  spinDelta =  _spinDelta

  method inverse()  =   
                        _transDelta#oppose();
                        _rotDelta#oppose();
                        _spinDelta#oppose();
                        _transSpeed#oppose();
                        _rotSpeed#oppose();
                        _spinSpeed#oppose()
  
  method settransSpeed x = _transSpeed <- x
  method setrotSpeed   x = _rotSpeed   <- x
  method setspinSpeed  x = _spinSpeed  <- x
  method settransDelta x = _transDelta <- x
  method setrotDelta   x = _rotDelta   <- x
  method setspinDelta  x = _spinDelta  <- x
  
end


(*********)
(* Scne *)
(*********)
class scene =
object(self)
  (* Dtermine si la scne est affiche ou non *)
  val mutable shown   = false
  (* Liste des objets de la scne *)
  val mutable contenu = (ref []:drawable list ref)
  (* Nom de la scne               *)
  (* "" si la scne n'a pas de nom *)
  val mutable name    = ""

  (*if the scene is an anonymous scene (indicated by the keyword "temporary")*)
  val mutable anonym = false

  (* Mthodes pour accder et modifier les attributs de la scne *)
  
  (*Draw the content of a scene*)
  method draw = List.iter (fun drawab -> drawab#draw ())  !contenu ;

  (*Compute the bounding box of the objects of the scene*)
  (*method getBounds = if (contenu= ref []) then (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
                     else boundingList !contenu*)
  (**)
  (*method drawObject objet = *)

  method show                  = shown <- true
  method hide                  = shown <- false
  method isShown               = shown
  method getName               = name
  method setName newName       = name <- newName
  method getAnonym             = anonym
  method setAnonym newVal      = anonym <- newVal
  method getContenu            = !contenu
  method setContenu newContenu = contenu <- ref newContenu
end

let dummyscene = new scene

(*************)
(* Commandes *)
(*************)
type commande =
    (* On affiche les scnes contenues dans la liste *)
  | Show     of scene list
    (* On cache les scnes contenues dans la liste *)
  | Hide     of scene list
    (* On substitue l'affichage de scnes par d'autre.     *)
    (* None -> toutes les scnes affiches sont remplaces *)
  | Replace  of scene list option * scene list
    (* Modification de la vue *)
  | View     of viewer * center * vertical
    (* On lance une commande shell *)
  | Command  of string
    (* On effectue un mouvement *)
  | Movement of mouvement
    (* Pause jusqu' ce que l'on appuie sur ESPACE *)
  | Pause



