open Types
open Parametres
open Triplet
open Draw
open Usefull

(************************************************)
(* Fixe la normale  norm ou  son oppos selon *)
(* drawInt.                                     *)
(************************************************)
(* Type : float * float * float -> bool -> unit *)
(************************************************)
let normeSens norm drawInt = GlDraw.normal3 (if drawInt
					     then (Triplet.oppose_triplet norm)
					     else norm)

(*************************************************************************************)
(** Central class for the display of scenes with all necessary parametres for OpenGL *)
(*************************************************************************************)
class viewport =
  object(self)
    
    (*---------*)
    (* Members *)
    (*---------*)

    (** List of currently displayed scenes *)
    val mutable sceneList = (ref[] : Types.scene list ref )

    
    (** Liste des objets transparents. *)
    val mutable blendFunctions = (ref []:((unit -> unit)*float) list ref)

    (******************)
    (** Axes, grilles *)
    (******************)

    (** Affichage ou non des axes *)
    val mutable affiche_axe    = ref false


    (** Axes formants le plan de la grille *)
    (*  Un plan invalide provode le        *)
    (*  non-affichage de la grille         *)
    val mutable grille1 = ref AxeNone
    val mutable grille2 = ref AxeNone

    (*********************)
    (** Mode d'affichage *)
    (*********************)

    (** Type d'affichage des polygones  (Fill, Line ou Point) *)
    val mutable polygon_mode = ref (`fill:[`fill | `line | `point])

    (** Eclair ou non *)
    val mutable lighting = ref true

    (** Brouillard ou non *)
    val mutable brouillard = ref true

    (** Stereo ou non *)
    val mutable stereo = ref false

    (** Ecartement des yeux *)
    val mutable inter_yeux = ref 1.0



    (*****************)
    (** Point de vue *)
    (*****************)


    (** Vecteurs orientant la vue :
         vectX : tel que (vectX, vectY, vectZ) orthonorme
         vectY : dfinie la verticale
         vectZ : vecteur unitaire porte par l'axe optique
    *)
    val mutable vectX = new point 1.0 0.0 0.0
    val mutable vectY = new point 0.0 1.0 0.0
    val mutable vectZ = new point 0.0 0.0 1.0


    (** Point objerv *)
    val mutable centre = new point 0.0 0.0 0.0


    (** Milieu du dessin *)
    val mutable milieu_dessin = new point 0.0 0.0 0.0


    (** Focus : distance entre la camera et le point observ *)
    val mutable zoom = ref 10.0


    (** Matrice de rotation du dessin sur lui-mme *)
    val mutable spin = ref ([| [| 1.0 ; 0.0 ; 0.0 ; 0.0 |];
			       [| 0.0 ; 1.0 ; 0.0 ; 0.0 |];
			       [| 0.0 ; 0.0 ; 1.0 ; 0.0 |];
			       [| 0.0 ; 0.0 ; 0.0 ; 1.0 |]
			    |])

    (** Taille du dessin *)
    val mutable rayon = ref 0.0

    (*-------*)
    (*Methods*)
    (*-------*)
								
    (************************)
    (** SETTERS AND GETTERS *)
    (************************)

    method setGrille1 newGrille1 = grille1 <- ref newGrille1

    method setGrille2 newGrille2 = grille2 <- ref newGrille2

    method setPolygonMode newMode = polygon_mode <- ref newMode
    method getPolygonMode = !polygon_mode

    method setLighting newLighting = lighting <- ref newLighting
    method getLighting = !lighting

    method setBrouillard newBrouillard = brouillard <- ref newBrouillard
    method getBrouillard = !brouillard

    method getCentre = centre

    method getMilieuDessin = milieu_dessin
    method setMilieuDessin newMilieu = milieu_dessin <- newMilieu

    method setZoom newZoom = zoom <- ref newZoom
    method getZoom = !zoom

    method getVectX = vectX

    method getVectY = vectY

    method getVectZ = vectZ

    method setSpin newSpin = spin <- ref newSpin

    method getRayon = !rayon
    method setRayon newRayon = rayon <- ref newRayon

    method setAfficheAxe newAfficheAxe = affiche_axe <- ref newAfficheAxe
    method getAfficheAxe = !affiche_axe
   
    method setStereo newStereo = stereo <- ref newStereo
    method getStereo = !stereo

    method setInterYeux newInterYeux = inter_yeux <- ref newInterYeux
    method getInterYeux = ! inter_yeux


    (******************************************************************)
    (** Insre dans lst un objet suivant sa profondeur z dcroissante.
        lst est une liste de mme sorte que blendObj.                 *)                 
    (******************************************************************)
    (* Type : (Types.primitive * Types.couleur * Types.draw) as `a -> *)
    (*        float -> GlMat.t -> (`a * float * GlMat.t) list ->      *)
    (*        (objet * float * GlMat.t) list                          *)
    (******************************************************************)
    method private insertOrdonne func z lst = match lst with
                                                | []   -> (func, z)::[]
					        | t::q -> let (_, tZ) = t
						          in
						            if tZ >= z
							    then
							      t::(self#insertOrdonne func z q)
							    else
							      (func, z)::lst


    method addBlendFunction func z = blendFunctions := self#insertOrdonne func z !blendFunctions

    (*********************)
    (** DISPLAYED SCENES *)
    (*********************)    

    (** Add a scene to the list of scenes *)
    method addScene scn = (*Check if scene not already exists*)
                          if not(List.exists (fun s-> s=scn) !sceneList) then
                            sceneList := scn::!sceneList;
                          (* JUST TEMPORARY *)
                          (*let t = (GMain.Timeout.add ~ms:50
                                	     ~callback:(fun () -> self#draw ();
					                          true)) in ()*)

    method addSceneList sceneList = List.iter (self#addScene) !sceneList

    (** Remove a scene from the list of scenes *)
    method removeScene scn = sceneList := List.filter (fun elem -> not (elem = scn)) !sceneList
    method removeSceneList sceneList = List.iter (self#removeScene) !sceneList

    (** Delete all scenes *)
    method clearScenes = sceneList <- (ref[] : Types.scene list ref )

    (** Draw all scenes *)
    method drawScenes = List.iter (fun scn -> scn#draw ) !sceneList

    (** Returns the list of currently displayed scenes *)
    method getScenes = !sceneList

    (** Returns the list of currently displayed named scenes *)
    method getNamedScenes = fold (fun scn -> scn) 
			         (fun scn tail -> if scn#getAnonym then tail
				                  else scn::tail
				 )
			         []
			         !sceneList



    (***********************)
    (** CAMERA ORIENTATION *)
    (***********************)
    
    (****************************************)
    (** Redfinit la matrice de perspective *)
    (****************************************)
    (* Type : unit -> unit                  *)
    (****************************************)
    method set_perspective () = let w = !Parametres.width
                                and h = !Parametres.height
                                and distance = !zoom +. (scalaire vectZ#getAll (diff_triplet centre#getAll milieu_dessin#getAll))
                                and r = !rayon *. Parametres.sqrt2
                                in
                                  let near = max 5.0 (distance -. r)
				  in
				    GlMat.mode `projection;
				    GlMat.load_identity ();
				    GluMat.perspective ~fovy:45.0 ~aspect:(w/.h) ~z:(near, distance +. r);
				    GlMat.mode `modelview


    (**********************************************************)
    (** Multipline la matrice spin (fichier parametres.ml) par 
        la matrice de rotation d'angle theta et d'axe vect     
        angle est en degrs.                                  *)
    (**********************************************************)
    (* Type : float * float * float -> float -> unit          *)
    (**********************************************************)
    method mult_spin vect theta = let angle = theta *. Parametres.pi /. 180.0
                                  in
                                    let c = cos angle
				    and s = sin angle
				    and (x, y, z) = normalise_triplet vect
				    in
				      spin := mult_matrice !spin [| [| x*.x*.(1.-.c)+.c    ; x*.y*.(1.-.c)-.z*.s ; x*.z*.(1.-.c)+.y*.s ; 0. |];
								    [| y*.x*.(1.-.c)+.z*.s ; y*.y*.(1.-.c)+.c    ; y*.z*.(1.-.c)-.x*.s ; 0. |];
								    [| x*.z*.(1.-.c)-.y*.s ; y*.z*.(1.-.c)+.x*.s ; z*.z*.(1.-.c)+.c    ; 0. |];
								    [| 0.                  ; 0.                  ; 0.                  ; 1. |]
								 |]



    (*****************************************************)
    (** Rotation de la camera autour du centre de l'cran.  
        X, y, z sont les angles en degrs de la rotation  
        autour des axes du repre de la camera           *)
    (*****************************************************)
    (* Type : float * float * float -> unit              *)
    (*****************************************************)
    method rotation (x, y, z) = if x <> 0.0
                                then rotate_vect vectY vectZ (Parametres.pi *. x /. 180.);
                                if y <> 0.0
				then rotate_vect vectZ vectX (Parametres.pi *. y /. 180.);
				if z <> 0.0
				then rotate_vect vectX vectY (Parametres.pi *. z /. 180.)



    (****************************************************)
    (** Rotation du dessin sur lui-mme                 
       x, y, z sont les angles en degrs de la rotation 
       autour des axes du repre de la camera.          
       x0, y0, z0 sont les angles en degrs de la       
       rotation autour des axes du repre absolu.       *)
    (****************************************************)
    (* Type : float * float * float ->                  *)
    (*        float * float * float -> unit             *)
    (****************************************************)
    method spin (x, y, z) (x0, y0, z0) = if  x <> 0.0
                                         then self#mult_spin vectX#getAll     x;
                                         if  y <> 0.0
					 then self#mult_spin vectY#getAll     y;
					 if  z <> 0.0
					 then self#mult_spin vectZ#getAll     z;
					 if x0 <> 0.0
					 then self#mult_spin (1.0, 0.0, 0.0) x0;
					 if y0 <> 0.0
					 then self#mult_spin (0.0, 1.0, 0.0) y0;
					 if z0 <> 0.0
					 then self#mult_spin (0.0, 0.0, 1.0) z0



    (*************************************************)
    (** Translation de la camera de vecteur (x, y, z)   
       dans le repre de la camera                   *)
    (*************************************************)
    (* Type : float * float * float -> unit          *)
    (*************************************************)
    method translation (x, y, z) = let trans = add_triplet (mult_triplet (vectX#getAll) x) (add_triplet (mult_triplet (vectY#getAll) y) (mult_triplet (vectZ#getAll) z))
                                   in
				     centre#updateAll trans


    (************)
    (** DRAWING *)
    (************)

    (******************************************)
    (** Dessine les axes de longueur 2*taille *)
    (******************************************)
    (* Type : float -> unit                   *)
    (******************************************)
    method private draw_axes taille = (* axe X *)
                                      GlDraw.color (1.0, 0.0, 0.0);
                                      GlDraw.begins `lines;
				      GlDraw.vertex3 (-. taille,       0.0,       0.0);
				      GlDraw.vertex3 (   taille,       0.0,       0.0);
				      (* axe Y *)
				      GlDraw.color (0.0, 1.0, 0.0);
				      GlDraw.vertex3 (      0.0, -. taille,       0.0);
				      GlDraw.vertex3 (      0.0,    taille,       0.0);
				      (* axe Z *)
				      GlDraw.color (0.0, 0.0, 1.0);
				      GlDraw.vertex3 (      0.0,       0.0, -. taille);
				      GlDraw.vertex3 (      0.0,       0.0,    taille);
				      GlDraw.ends ()

    (*****************************************)
    (** Dessine le nom des axes de longueur.  
        taille est la demi-longueur des axes.*)
    (*****************************************)
    (* Type : float -> unit                  *)
    (*****************************************)
    method draw_axes_name taille = Gl.enable `blend;
                                   GlPix.raster_pos ~x:taille ~y:1.0 ~z:0.0 ();
				   GlPix.draw plusX;
				   
				   GlPix.raster_pos ~x:(-.taille) ~y:1.0 ~z:0.0 ();
				   GlPix.draw moinsX;
				   
				   GlPix.raster_pos ~x:1.0 ~y:taille ~z:0.0 ();
				   GlPix.draw plusY;
				   
				   GlPix.raster_pos ~x:1.0 ~y:(-.taille) ~z:0.0 ();
				   GlPix.draw moinsY;
				   
				   GlPix.raster_pos ~x:0.0 ~y:1.0 ~z:taille ();
				   GlPix.draw plusZ;
				   
				   GlPix.raster_pos ~x:0.0 ~y:1.0 ~z:(-.taille) ();
				   GlPix.draw moinsZ;

				   Gl.disable `blend

    (*********************************)
    (** Dessine la grille sur le plan 
        form des axes axis1 et axis2*)
    (*********************************)
    (* Type : axe -> axe -> unit     *)
    (*********************************)
    method draw_grille axis1 axis2 = let limite = floor !rayon
                                     in
                                       let i = ref (-.limite)
				       in
				         let axis_func = match (axis1, axis2)
					   with
					     | (AxeX, AxeY) | (AxeY, AxeX) -> (function i -> GlDraw.vertex3 (-. !rayon,        i,      0.0);
						                                             GlDraw.vertex3 (   !rayon,        i,      0.0);
											     GlDraw.vertex3 (        i,-. !rayon,      0.0);
											     GlDraw.vertex3 (        i,   !rayon,      0.0))
					     | (AxeY, AxeZ) | (AxeZ, AxeY) -> (function i -> GlDraw.vertex3 (      0.0,-. !rayon,        i);
						                                             GlDraw.vertex3 (      0.0,   !rayon,        i);
											     GlDraw.vertex3 (      0.0,        i,-. !rayon);
											     GlDraw.vertex3 (      0.0,        i,   !rayon))
					     | (AxeZ, AxeX) | (AxeX, AxeZ) -> (function i -> GlDraw.vertex3 (        i,      0.0,-. !rayon);
						                                             GlDraw.vertex3 (        i,      0.0,   !rayon);
											     GlDraw.vertex3 (-. !rayon,      0.0,        i);
											     GlDraw.vertex3 (   !rayon,      0.0,        i))
					     | _                            -> ignore
					 in
					   GlDraw.color (0.7, 0.7, 0.7);
					   GlDraw.begins `lines;
					   
					   while !i < limite
					   do
					     axis_func !i;
					     i := !i +. 1.0
					   done;
					   GlDraw.ends ()


    (****************************************)
    (** Dessine toutes les scnes depuis une 
        camera excentre de l'axe de delta  *)
    (****************************************)
    (* Type : float -> unit                 *)
    (****************************************)
    method view delta =  (* On met  jour la matrice de projection *)
                         self#set_perspective ();
                         GlMat.load_identity ();
			 (* On place la lumire *)
			 GlLight.light ~num:0 (`position (*10.0, 10.0, 10.0, 0.0*)
						  (*let (lposX,lposY,lposZ) = add_triplet
						    (add_triplet
						    Parametres.centre#getAll
						       (mult_triplet
						    (Parametres.vectZ#getAll)
						    !Parametres.zoom
						    )
						    )
						    (mult_triplet Parametres.vectX#getAll delta)
						    in
						    (lposX, lposY, lposZ, 0.0)*)
						  (0.0, 0.0, !zoom, 0.0)
					      );

			 (* On place la camera *)
			 GluMat.look_at ~eye:(add_triplet
						 (add_triplet
						     centre#getAll
						     (mult_triplet
							 (vectZ#getAll)
							 !zoom
						     )
						 )
					      (mult_triplet vectX#getAll delta)
					     )
			   ~center:centre#getAll
			   ~up:vectY#getAll;
			 
			 (* Rotation de la scne sur elle-mme *)
			 GlMat.translate3 milieu_dessin#getAll;
			 GlMat.mult (GlMat.of_array !spin);
			 GlMat.translate3 (mult_triplet milieu_dessin#getAll (-1.0));
			 
			 (* Affichage de la grille et des axes si ncessaire *)
			 self#draw_grille !grille1 !grille2;
			 if !affiche_axe
			 then self#draw_axes !rayon;
			 
			 (* Acitivation du brouillard et de la lumire si ncessaire pour les scnes *)
			 if !lighting
			 then Gl.enable `lighting;
			 if !brouillard
			 then Gl.enable `fog;
			 
			 (* Dessin les scnes *)
			 self#drawScenes;

			 (* Dessine les objets transparents *)
			 let sbt () = GlFunc.depth_mask false;
			   (*List.iter (fun (obj, _, mat) -> GlMat.load mat;
			                                   obj#draw();
							   obj#setDrawBlend false;
				     )
			             !blendObj;*)
			              List.iter (fun (func,z) -> func () )
					        !blendFunctions;
			   
			   GlFunc.depth_mask true
			 in
			   if !Parametres.isVectoring
			   then
			     begin
			       Vectoriel.psEnable `blend;
			       sbt ();
			       Vectoriel.psDisable `blend
			     end
			   else
			     begin
			       Gl.enable `blend;
			       sbt ();
			       Gl.disable `blend
			     end;
			   blendFunctions := [];
			   
			   if !affiche_axe then self#draw_axes_name !rayon;
			   
			   Gl.disable `lighting;
			   Gl.disable `fog
			     


    (*****************************************)
    (** Dessine toutes les scenes suivant si 
        le mode stro est activ ou non.    *)
    (*****************************************)
    (* Type : unit -> unit                   *)
    (*****************************************)
    method draw () =  (*Avoid parallel calls of function*)
                      Mutex.lock drawMutex;
                      if !stereo
		      then
			begin
			  (*Mode stro *)
			  GlFunc.draw_buffer `back;
			  GlClear.clear [`color; `depth];
			  (* Oeil droit *)
			  GlFunc.draw_buffer `back_right;
			  self#view !inter_yeux;
			  (* Oeil gauche *)
			  GlFunc.draw_buffer `back_left;
			  GlClear.clear [`depth];
			  self#view (-. !inter_yeux)
			end
		      else
			begin
			  (* Mode "mono" *)
			  GlClear.clear [`color; `depth];
			  self#view 0.0
			end;
		      !Parametres.swapbuffers ();
		      Mutex.unlock drawMutex
			
  end

let view = new viewport
